import random
import threading
import time
import warnings
from dataclasses import asdict, dataclass
from itertools import islice
from queue import Queue
from typing import (
    Any,
    Callable,
    Deque,
    Dict,
    Generator,
    Iterable,
    List,
    Literal,
    Optional,
    Sequence,
    Tuple,
    Union,
)

import numpy as np
import torch

from lhotse import CutSet, Seconds
from lhotse.cut import Cut
from lhotse.dataset.dataloading import resolve_seed
from lhotse.dataset.sampling.base import (
    CutSampler,
    EpochDiagnostics,
    SamplingConstraint,
    SamplingDiagnostics,
    TimeConstraint,
)
from lhotse.dataset.sampling.dynamic import DurationBatcher, Filter, check_constraint
from lhotse.utils import ifnone


class DynamicBucketingSampler(CutSampler):
    """
    A dynamic (streaming) variant of :class:`~lhotse.dataset.sampling.bucketing.BucketingSampler`,
    that doesn't require reading the whole cut set into memory.

    The basic idea is to sample N (e.g. ~10k) cuts and estimate the boundary durations for buckets.
    Then, we maintain a buffer of M cuts (stored separately in K buckets) and every time we sample a batch,
    we consume the input cut iterable for the same amount of cuts.
    The memory consumption is limited by M at all times.

    For scenarios such as ASR, VAD, Speaker ID, or TTS training, this class supports single CutSet
    iteration. Example::

        >>> cuts = CutSet(...)
        >>> sampler = DynamicBucketingSampler(cuts, max_duration=100)
        >>> for batch in sampler:
        ...     assert isinstance(batch, CutSet)

    For other scenarios that require pairs (or triplets, etc.) of utterances, this class supports
    zipping multiple CutSets together. Such scenarios could be voice conversion, speech translation,
    contrastive self-supervised training, etc. Example::

        >>> source_cuts = CutSet(...)
        >>> target_cuts = CutSet(...)
        >>> sampler = DynamicBucketingSampler(source_cuts, target_cuts, max_duration=100)
        >>> for batch in sampler:
        ...     assert isinstance(batch, tuple)
        ...     assert len(batch) == 2
        ...     assert isinstance(batch[0], CutSet)
        ...     assert isinstance(batch[1], CutSet)

    .. note:: for cut pairs, triplets, etc. the user is responsible for ensuring that the CutSets
        are all sorted so that when iterated over sequentially, the items are matched.
        We take care of preserving the right ordering internally, e.g., when shuffling.
        By default, we check that the cut IDs are matching, but that can be disabled.

    .. caution:: when using :meth:`DynamicBucketingSampler.filter` to filter some cuts with more than
        one CutSet to sample from, we sample one cut from every CutSet, and expect that all of the cuts
        satisfy the predicate -- otherwise, they are all discarded from being sampled.
    """

    def __init__(
        self,
        *cuts: Iterable,
        max_duration: Optional[Seconds] = None,
        max_cuts: Optional[int] = None,
        constraint: Optional[SamplingConstraint] = None,
        num_buckets: Optional[int] = 10,
        shuffle: bool = False,
        drop_last: bool = False,
        consistent_ids: bool = True,
        duration_bins: List[Seconds] = None,
        num_cuts_for_bins_estimate: int = 10000,
        buffer_size: int = 20000,
        quadratic_duration: Optional[Seconds] = None,
        world_size: Optional[int] = None,
        rank: Optional[int] = None,
        seed: Union[int, Literal["randomized", "trng"]] = 0,
        sync_buckets: bool = True,
        concurrent: bool = False,
        strict=None,
        shuffle_buffer_size=None,
    ) -> None:
        """
        :param cuts: one or more CutSets (when more than one, will yield tuples of CutSets as mini-batches)
        :param max_duration: The maximum total recording duration from ``cuts``.
            Note: with multiple CutSets, ``max_duration`` constraint applies only to the first CutSet.
        :param max_cuts: The maximum total number of ``cuts`` per batch.
            When only ``max_duration`` is specified, this sampler yields static batch sizes.
        :param num_buckets: how many buckets to create. Ignored if duration_bins are provided.
        :param shuffle: When ``True``, the cuts will be shuffled dynamically with
            a reservoir-sampling-based algorithm.
            Convenient when mini-batch loop is inside an outer epoch-level loop, e.g.:
            `for epoch in range(10): for batch in dataset: ...` as every epoch will see a
            different cuts order.
        :param drop_last: When ``True``, we will drop all incomplete batches.
            A batch is considered incomplete if it depleted a bucket before
            hitting the constraint such as max_duration, max_cuts, etc.
        :param consistent_ids: Only affects processing of multiple CutSets.
            When ``True``, at each sampling step we check cuts from all CutSets have the same ID
            (i.e., the first cut from every CutSet should have the same ID, same for the second, third, etc.).
        :param duration_bins: A list of floats (seconds); when provided, we'll skip the initial
            estimation of bucket duration bins (useful to speed-up the launching of experiments).
        :param num_cuts_for_bins_estimate: We will draw this many cuts to estimate the duration bins
            for creating similar-duration buckets.
            Larger number means a better estimate to the data distribution, possibly at a longer init cost.
        :param buffer_size: How many cuts (or cut pairs, triplets) we hold at any time across all
            of the buckets.
            Increasing ``max_duration`` (batch_size) or ``num_buckets`` might require increasing this number.
            Larger number here will also improve shuffling capabilities.
            It will result in larger memory usage.
        :param quadratic_duration: When set, it adds an extra penalty that's quadratic in size w.r.t.
            a cuts duration. This helps get a more even GPU utilization across different input lengths
            when models have quadratic input complexity. Set between 15 and 40 for transformers.
        :param sync_buckets: When set, we'll try to make each DDP rank sample from as close
            duration buckets as possible to minimize the tail worker effect.
        :param concurrent: Enabling concurrency eliminates most of the waiting to pre-populate the
            bucketing buffers before the sampler starts yielding examples. For tarred/Lhotse Shar data
            this can speed up the start of the training. Note that enabling concurrency will cause the
            sampling results to be non-deterministic. This feature is experimental.
        :param world_size: Total number of distributed nodes. We will try to infer it by default.
        :param rank: Index of distributed node. We will try to infer it by default.
        :param seed: Random seed used to consistently shuffle the dataset across different processes.
        """
        super().__init__(
            drop_last=drop_last, world_size=world_size, rank=rank, seed=seed
        )
        if not all(cs.is_lazy for cs in cuts if isinstance(cs, CutSet)):
            warnings.warn(
                "You are using DynamicBucketingSampler with an eagerly read CutSet. "
                "You won't see any memory/speed benefits with that setup. "
                "Either use 'CutSet.from_jsonl_lazy' to read the CutSet lazily, or use a BucketingSampler instead."
            )
        self.cuts = cuts
        self.max_duration = max_duration
        self.max_cuts = max_cuts
        self.constraint = constraint
        self.shuffle = shuffle
        self.consistent_ids = consistent_ids
        self.num_cuts_for_bins_estimate = num_cuts_for_bins_estimate
        self.buffer_size = buffer_size
        self.quadratic_duration = quadratic_duration
        self.sync_buckets = sync_buckets
        self.concurrent = concurrent
        self.rng = None
        check_constraint(constraint, max_duration, max_cuts)

        if strict is not None:
            warnings.warn(
                "In Lhotse v1.4 all samplers act as if 'strict=True'. "
                "Sampler's argument 'strict' will be removed in a future Lhotse release.",
                category=DeprecationWarning,
            )

        if shuffle_buffer_size is not None:
            _emit_shuffle_buffer_size_warning()
            self.buffer_size += shuffle_buffer_size

        if duration_bins is not None:
            assert list(duration_bins) == sorted(
                duration_bins
            ), "Duration bins must be sorted ascendingly."
            self.duration_bins = duration_bins
            self.num_buckets = len(duration_bins) + 1
        else:
            if constraint is None:
                constraint = TimeConstraint(
                    max_duration=self.max_duration,
                    max_cuts=self.max_cuts,
                    quadratic_duration=self.quadratic_duration,
                )
            self.duration_bins = estimate_duration_buckets(
                islice(self.cuts[0], num_cuts_for_bins_estimate),
                num_buckets=num_buckets,
                constraint=constraint,
            )

    def state_dict(self) -> Dict[str, Any]:
        assert (
            self.constraint is None
        ), "state_dict() is not supported with samplers that use a custom constraint."
        sd = super().state_dict()
        sd.update(
            {
                "max_duration": self.max_duration,
                "max_cuts": self.max_cuts,
                "consistent_ids": self.consistent_ids,
                "buffer_size": self.buffer_size,
                "num_cuts_for_bins_estimate": self.num_cuts_for_bins_estimate,
                "quadratic_duration": self.quadratic_duration,
            }
        )
        return sd

    def load_state_dict(self, sd: Dict[str, Any]) -> None:
        self.max_duration = sd.pop("max_duration")
        self.max_cuts = sd.pop("max_cuts")
        self.consistent_ids = sd.pop("consistent_ids")
        self.num_cuts_for_bins_estimate = sd.pop("num_cuts_for_bins_estimate")
        self.buffer_size = sd.pop("buffer_size")
        if "shuffle_buffer_size" in sd:
            _emit_shuffle_buffer_size_warning()
            shuffle_buffer_size = sd.pop("shuffle_buffer_size")
            self.buffer_size += shuffle_buffer_size
        self.quadratic_duration = sd.pop("quadratic_duration", None)
        sd.pop("strict", None)  # backward compatibility
        super().load_state_dict(sd)
        self._fast_forward()

    def _fast_forward(self):
        current_epoch = self.diagnostics.current_epoch
        num_batches_to_iter = self.diagnostics.current_epoch_stats.total_batches

        # Set the right epoch
        self.set_epoch(current_epoch)
        # Reset diagnostics for this epoch as we're about to re-iterate
        self.diagnostics.stats_per_epoch[current_epoch] = EpochDiagnostics(
            epoch=current_epoch
        )

        self._just_restored_state = False
        iter(self)
        for _ in range(num_batches_to_iter):
            next(self)
        self._just_restored_state = True

    def __iter__(self) -> "DynamicBucketingSampler":
        if self._just_restored_state:
            return self
        seed = resolve_seed(self.seed)
        self.rng = random.Random(seed + self.epoch)
        if self.sync_buckets:
            # Bucket sync requested. To achieve that we will fix the RNG seed for a special bucket RNG
            # in a deterministic way. We also consider whether the sampler object lives in the training loop
            # process (map-style dataset or num_workers=0) or the dataloading subprocess (iterable-style dataset).
            # In the latter case, we want each worker to choose different buckets but still be in sync
            # with workers with the same IDs on other ranks.
            # Note: PyTorch dataloader always iterates workers sequentially, so they won't get out-of-order.
            bucket_rng_seed = 1234
            worker_info = torch.utils.data.get_worker_info()
            if worker_info is not None:
                bucket_rng_seed += worker_info.id
            bucket_rng = random.Random(bucket_rng_seed)
        else:
            bucket_rng = None
        # Why reset the current epoch?
        # Either we are iterating the epoch for the first time and it's a no-op,
        # or we are iterating the same epoch again, in which case setting more steps
        # than are actually available per epoch would have broken the checkpoint restoration.
        self.diagnostics.reset_current_epoch()
        # Initiate iteration
        cuts_iter = [iter(cs) for cs in self.cuts]
        # Apply filter predicate
        cuts_iter = Filter(
            iterator=zip(*cuts_iter),
            predicate=lambda tpl: all(self._filter_fn(c) for c in tpl),
            diagnostics=self.diagnostics,
        )
        # Convert Iterable[Cut] -> Iterable[CutSet]
        cuts_iter = DynamicBucketer(
            cuts_iter,
            duration_bins=self.duration_bins,
            world_size=self.world_size,
            max_duration=self.max_duration,
            max_cuts=self.max_cuts,
            constraint=self.constraint,
            drop_last=self.drop_last,
            buffer_size=self.buffer_size,
            quadratic_duration=self.quadratic_duration,
            shuffle=self.shuffle,
            rng=self.rng,
            bucket_rng=bucket_rng,
            concurrent=self.concurrent,
            diagnostics=self.diagnostics,
        )
        self.cuts_iter = iter(cuts_iter)
        return self

    def _next_batch(self) -> Union[CutSet, Tuple[CutSet]]:
        batch = next(self.cuts_iter)
        if self.consistent_ids and isinstance(batch, tuple):
            for cuts in zip(*batch):
                expected_id = cuts[0].id
                assert all(c.id == expected_id for c in cuts[1:]), (
                    f"The input CutSet are not sorted by cut ID in the same way. "
                    f"We sampled the following mismatched cut IDs: {', '.join(c.id for c in cuts)}. "
                    f"If this is expected, pass the argument 'consistent_ids=False' to DynamicBucketingSampler."
                )
        return batch

    @property
    def remaining_duration(self) -> Optional[float]:
        return None

    @property
    def remaining_cuts(self) -> Optional[int]:
        return None

    @property
    def num_cuts(self) -> Optional[int]:
        return None


@dataclass
class FixedBucketBatchSizeConstraint(SamplingConstraint):
    """
    Sampling constraint that accepts a pre-defined batch size for each bucket.
    It uses the example's sequence length to determine which bucket we're sampling for,
    and otherwise the batch size is locally static for each bucket.

    This constraint doesn't support samples longer than the upper bound of the last bucket;
    if such sample is provided, we will raise an exception.
    """

    max_seq_len_buckets: List[float]
    batch_sizes: List[int]
    current_bucket: Union[int, None] = None
    num_cuts: int = 0

    def __post_init__(self):
        assert sorted(self.max_seq_len_buckets) == list(self.max_seq_len_buckets)

    def is_active(self) -> bool:
        return True

    def add(self, example: Cut) -> None:
        """
        Increment the internal counter for the time constraint,
        selecting the right property from the input ``cut`` object.
        """
        seqlen = self.measure_length(example)
        bucket_idx = self.select_bucket(
            buckets=self.max_seq_len_buckets, example_len=seqlen
        )
        assert bucket_idx < len(self.max_seq_len_buckets), (
            f"Received example with sequence length {seqlen} that exceeds "
            f"the highest allowed length {self.max_seq_len_buckets[-1]}."
        )
        if self.current_bucket is None:
            self.current_bucket = bucket_idx
        else:
            assert self.current_bucket == bucket_idx, (
                f"User error: FixedBucketBatchSizeConstraint is supposed to be used only on one bucket. "
                f"The example we received has sequence length {seqlen} which is outside of the allowed bounds "
                f"for bucket index {bucket_idx} in buckets {self.max_seq_len_buckets}."
            )
        self.num_cuts += 1

    def exceeded(self) -> bool:
        """Is the constraint exceeded or not."""
        return self.num_cuts > self.batch_sizes[self.current_bucket]

    def close_to_exceeding(self) -> bool:
        """
        Check if the batch is close to satisfying the constraints.
        We define "closeness" as: if we added one more cut that has
        duration/num_frames/num_samples equal to the longest seen cut
        in the current batch, then the batch would have exceeded the constraints.
        """
        return self.num_cuts >= self.batch_sizes[self.current_bucket]

    def reset(self) -> None:
        """
        Reset the internal counter (to be used after a batch was created,
        to start collecting a new one).
        """
        self.current_bucket = None
        self.num_cuts = 0

    def measure_length(self, example: Cut) -> float:
        return example.duration

    def state_dict(self) -> Dict[str, Any]:
        return asdict(self)

    def load_state_dict(self, state_dict: Dict[str, Any]) -> None:
        self.max_seq_len_buckets = state_dict.pop("max_seq_len_buckets")
        self.batch_sizes = state_dict.pop("batch_sizes")
        self.current_bucket = state_dict.pop("current_bucket")
        self.num_cuts = state_dict.pop("num_cuts")
        assert len(state_dict) == 0, (
            "Error in FixedBucketBatchSizeConstraint.load_state_dict(): Unexpected keys:\n- "
            + "\n- ".join(state_dict.keys())
        )

    def __add__(
        self, other: "FixedBucketBatchSizeConstraint"
    ) -> "FixedBucketBatchSizeConstraint":
        for key in ("max_seq_len_buckets", "batch_sizes", "current_bucket"):
            self_attr = getattr(self, key)
            other_attr = getattr(other, key)
            is_none = self_attr is None and other_attr is None
            assert is_none or self_attr == other_attr, (
                f"To add two TimeConstraint objects, they need to represent the same constraint "
                f"(got self.{key}={self_attr} != other.{key}={other_attr})."
            )
        return FixedBucketBatchSizeConstraint(
            max_seq_len_buckets=self.max_seq_len_buckets,
            batch_sizes=self.batch_sizes,
            current_bucket=self.current_bucket,
            num_cuts=self.num_cuts + other.num_cuts,
        )

    def __eq__(self, other: "TimeConstraint") -> bool:
        return (
            isinstance(other, FixedBucketBatchSizeConstraint)
            and self.max_seq_len_buckets == other.max_seq_len_buckets
            and self.batch_sizes == other.batch_sizes
            and self.current_bucket == other.current_bucket
        )


def estimate_duration_buckets(
    cuts: Iterable[Cut],
    num_buckets: int,
    constraint: Optional[SamplingConstraint] = None,
) -> List[float]:
    """
    Given an iterable of cuts and a desired number of buckets, select duration values
    that should start each bucket.

    The returned list, ``bins``, has ``num_buckets - 1`` elements.
    The first bucket should contain cuts with duration ``0 <= d < bins[0]``;
    the last bucket should contain cuts with duration ``bins[-1] <= d < float("inf")``,
    ``i``-th bucket should contain cuts with duration ``bins[i - 1] <= d < bins[i]``.

    :param cuts: an iterable of :class:`lhotse.cut.Cut`.
    :param num_buckets: desired number of buckets.
    :param constraint: object with ``.measure_length()`` method that's used to determine
        the size of each sample. If ``None``, we'll use ``TimeConstraint``.
    :return: a list of boundary duration values (floats).
    """
    assert num_buckets > 1

    if constraint is None:
        constraint = TimeConstraint()

    sizes = np.array([constraint.measure_length(c) for c in cuts])
    sizes.sort()
    assert num_buckets <= sizes.shape[0], (
        f"The number of buckets ({num_buckets}) must be smaller than "
        f"or equal to the number of cuts ({sizes.shape[0]})."
    )
    size_per_bucket = sizes.sum() / num_buckets

    bins = []
    tot = 0.0
    for size in sizes:
        if tot > size_per_bucket:
            bins.append(size)
            tot = 0.0
        tot += size

    return bins


class BucketSelectionState:
    """
    Helper class used in the context of bucket selection synchronization across DDP ranks.
    It's only necessary when using a map-style dataset (i.e., the sampler lives in the training loop process)
    and world_size is greater than 1. In these cases we have to use the same bucket idx ``world_size`` times
    to ensure each rank uses the same bucket. This is due to how CutSampler distributes mini-batches
    across ranks, ensuring the number of steps is always equal for each rank.
    """

    def __init__(
        self, bucket_rng: random.Random, num_buckets: int, world_size: int
    ) -> None:
        self._bucket_rng = bucket_rng
        self._num_buckets = num_buckets
        self._world_size = world_size
        self._usage_count = 0
        self._bucket_idx = None

    def select_bucket_idx(self) -> int:
        if self._bucket_idx is None or self._usage_count == self._world_size:
            self._bucket_idx = self._bucket_rng.randrange(self._num_buckets)
            self._usage_count = 0
        self._usage_count += 1
        return self._bucket_idx

    def save(self) -> Dict[str, Any]:
        return {
            "_bucket_rng": self._bucket_rng.getstate(),
            "_bucket_idx": self._bucket_idx,
            "_usage_count": self._usage_count,
        }

    def restore(self, ckpt: Dict[str, Any]) -> None:
        self._bucket_rng.setstate(ckpt["_bucket_rng"])
        self._bucket_idx = ckpt["_bucket_idx"]
        self._usage_count = ckpt["_usage_count"]


class DynamicBucketer:
    def __init__(
        self,
        cuts: Iterable[Union[Cut, Tuple[Cut]]],
        duration_bins: List[Seconds],
        world_size: int,
        max_duration: Optional[Seconds] = None,
        max_cuts: Optional[int] = None,
        constraint: Optional[SamplingConstraint] = None,
        drop_last: bool = False,
        buffer_size: int = 10000,
        quadratic_duration: Optional[Seconds] = None,
        shuffle: bool = False,
        rng: random.Random = None,
        bucket_rng: random.Random = None,
        concurrent: bool = False,
        diagnostics: Optional[SamplingDiagnostics] = None,
    ) -> None:
        self.cuts = cuts
        self.duration_bins = duration_bins
        self.world_size = world_size
        self.max_duration = max_duration
        self.max_cuts = max_cuts
        self.constraint = constraint
        self.drop_last = drop_last
        self.buffer_size = buffer_size
        self.quadratic_duration = quadratic_duration
        self.diagnostics = ifnone(diagnostics, SamplingDiagnostics())
        if rng is None:
            rng = random.Random()
        self.rng = rng
        self.bucket_rng = bucket_rng
        self.shuffle = shuffle
        self.concurrent = concurrent

        assert duration_bins == sorted(duration_bins), (
            f"Argument list for 'duration_bins' is expected to be in "
            f"sorted order (got: {duration_bins})."
        )
        check_constraint(constraint, max_duration, max_cuts)

        if self.constraint is None:
            self.constraint = TimeConstraint(
                max_duration=self.max_duration,
                max_cuts=self.max_cuts,
                quadratic_duration=self.quadratic_duration,
            )

        # A heuristic diagnostic first, for finding the right settings.
        if max_duration is not None:
            mean_duration = np.mean(duration_bins)
            expected_buffer_duration = buffer_size * mean_duration
            expected_bucket_duration = expected_buffer_duration / (
                len(duration_bins) + 1
            )
            if expected_bucket_duration < max_duration:
                warnings.warn(
                    f"Your 'buffer_size' setting of {buffer_size} might be too low to satisfy "
                    f"a 'max_duration' of {max_duration} (given our best guess)."
                )

        # Init: create empty buckets (note: `num_buckets = len(duration_bins) + 1`).
        self.buckets: List[Queue] = [Queue() for _ in range(len(duration_bins) + 1)]

        self._producer_thread = None

    def __iter__(self) -> Generator[CutSet, None, None]:
        # Init: sample `buffer_size` cuts and assign them to the right buckets.
        self.cuts_iter = iter(self.cuts)

        if self.concurrent:
            self._start_data_producer_thread()
            self._maybe_wait_for_producer()
        else:
            self._collect_cuts_in_buckets(self.buffer_size)

        state = BucketSelectionState(
            bucket_rng=self.bucket_rng,
            num_buckets=len(self.buckets),
            world_size=self.world_size,
        )

        # The iteration code starts here.
        # On each step we're sampling a new batch.
        try:
            while True:
                sampling_bucket = self._select_bucket(state)
                # Apply random shuffling if requested: we'll shuffle the items present within the bucket.
                maybe_shuffled = sampling_bucket
                indexes_used = []
                if self.shuffle:
                    maybe_shuffled = pick_at_random(
                        maybe_shuffled, rng=self.rng, out_indexes_used=indexes_used
                    )
                else:
                    with sampling_bucket.mutex:
                        maybe_shuffled = list(sampling_bucket.queue)
                # Sample one batch from that bucket and yield it to the caller.
                batcher = DurationBatcher(
                    maybe_shuffled,
                    constraint=self.constraint.copy(),
                    diagnostics=self.diagnostics,
                )
                batch = next(iter(batcher))
                if isinstance(batch, tuple):
                    batch_size = len(batch[0])
                else:
                    batch_size = len(batch)
                yield batch
                # Remove sampled cuts from the bucket.
                if indexes_used:
                    # Shuffling, sort indexes of yielded elements largest -> smallest and remove them
                    indexes_used.sort(reverse=True)
                    with sampling_bucket.mutex:
                        _q = sampling_bucket.queue
                        for idx in indexes_used:
                            del _q[idx]
                else:
                    # No shuffling, remove first N
                    for _ in range(batch_size):
                        sampling_bucket.get()
                # Fetch new cuts and add them to appropriate buckets.
                if self.concurrent:
                    self._maybe_wait_for_producer()
                else:
                    self._collect_cuts_in_buckets(batch_size)
        except StopIteration:
            pass

        # Cleanup.
        self.cuts_iter = None

    def _select_bucket(self, state: BucketSelectionState) -> Queue:
        if self.bucket_rng is None:
            # Bucket selection algo 1:
            # * there is just one RNG for choosing buckets and choosing samples randomly from the buckets
            # * check which buckets are ready, and then use the RNG to select one of them.
            # * no guarantees about bucket selection sync across GPUs.
            ready_buckets = [b for b in self.buckets if self._is_ready(b)]
            if not ready_buckets:
                # No bucket has enough data to yield for the last full batch.
                non_empty_buckets = [b for b in self.buckets if b]
                if self.drop_last or len(non_empty_buckets) == 0:
                    # Either the user requested only full batches, or we have nothing left.
                    raise StopIteration()
                else:
                    # Sample from partial batches that are left.
                    ready_buckets = non_empty_buckets
            # Choose a bucket to sample from.
            # We'll only select from the buckets that have a full batch available.
            return self.rng.choice(ready_buckets)
        else:
            # Bucket selection algo 2:
            # * bucket selection has its own independent RNG.
            # * when bucket selection RNG is initialized identically on all ranks/workers,
            #     then each rank will initially select the same bucket for batch sampling
            # * if one of the ranks selects a bucket that is not filled enough,
            #     it will scan the neighbouring buckets until it finds one that's ready
            # * if no bucket is ready, we end iteration

            def scan_buckets(predicate: Callable[[Queue], bool]) -> int:
                bucket_idx = state.select_bucket_idx()

                def valid_idx() -> bool:
                    return 0 <= bucket_idx < len(self.buckets)

                num_attempts = 0
                seen_min, seen_max = bucket_idx, bucket_idx
                while not (valid_idx() and predicate(self.buckets[bucket_idx])):
                    if seen_min < 0 and seen_max >= len(self.buckets):
                        raise BucketsDontHaveEnoughData()
                    num_attempts += 1
                    bucket_idx = (
                        bucket_idx + (1 if num_attempts % 2 == 0 else -1) * num_attempts
                    )
                    seen_min = min(seen_min, bucket_idx)
                    seen_max = max(seen_max, bucket_idx)

                return bucket_idx

            # This try/except is first trying to choose a bucket to sample a full mini-batch from,
            # and if that fails and drop_last=False, it tries again, this time accepting partial mini-batch.
            # Because we have no guarantee that samplers in different ranks will start exhausting the buckets
            # at the same time, it takes only a single occurrence of all buckets not being ready to permanently
            # run out-of-sync.
            # For this reason, we create a checkpoint of the bucket sampling state, and if we go into except
            # fallback, we restore this state first to ensure we use the bucket_rng exactly the same number
            # of times on each rank, no matter the circumstance.
            ckpt = state.save()
            try:
                # Typical case: at least one bucket has enough data to sample from.
                selected_bucket_idx = scan_buckets(self._is_ready)
            except BucketsDontHaveEnoughData:
                # We didn't hit the typical case either because we are finishing
                # the epoch, or because the buffers are too small.
                if self.drop_last:
                    # The user doesn't want partial mini-batches: early exit.
                    raise StopIteration()
                # The user wants to iterate the full dataset.
                # We'll try again, this time accepting buckets that have any amount of data available,
                # which may yield partial batches.
                try:
                    state.restore(ckpt)
                    selected_bucket_idx = scan_buckets(lambda b: b.qsize() > 0)
                except BucketsDontHaveEnoughData:
                    # We exhausted the full dataset.
                    raise StopIteration()

            return self.buckets[selected_bucket_idx]

    def _is_ready(self, bucket: Queue) -> bool:
        tot = self.constraint.copy()
        with bucket.mutex:
            contents = list(bucket.queue)
        for c in contents:
            tot.add(c[0] if isinstance(c, tuple) else c)
            if tot.close_to_exceeding():
                return True
        return False

    def _start_data_producer_thread(self):
        """Start concurrent filling of the bucket buffer in a background thread."""

        def producer():
            try:
                self._source_exhausted = False
                while not self._source_exhausted:
                    if sum(b.qsize() for b in self.buckets) == self.buffer_size:
                        time.sleep(0.1)
                        continue
                    cuts = next(self.cuts_iter)
                    bucket_idx = self.constraint.select_bucket(
                        buckets=self.duration_bins,
                        example=cuts[0] if isinstance(cuts, tuple) else cuts,
                    )
                    self.buckets[bucket_idx].put(cuts)
            except StopIteration:
                self._source_exhausted = True

        self._producer_thread = threading.Thread(target=producer)
        self._producer_thread.start()

    def _maybe_wait_for_producer(self):
        """Triggers wait for producer if the bucket buffers are less than 10% utilized."""
        while (
            sum(b.qsize() for b in self.buckets) < self.buffer_size / 10
            and not self._source_exhausted
        ):
            time.sleep(1.0)

    def _collect_cuts_in_buckets(self, n_cuts: int) -> None:
        """Fetches ``n_cuts`` from the input data iterable. Doesn't use concurrency."""
        try:
            for _ in range(n_cuts):
                cuts = next(self.cuts_iter)
                bucket_idx = self.constraint.select_bucket(
                    buckets=self.duration_bins,
                    example=cuts[0] if isinstance(cuts, tuple) else cuts,
                )
                self.buckets[bucket_idx].put(cuts)
        except StopIteration:
            pass

    def __del__(self):
        if (
            self.concurrent
            and self._producer_thread is not None
            and self._producer_thread.is_alive()
        ):
            self._source_exhausted = True
            self._producer_thread.join()


def pick_at_random(
    bucket: Queue,
    rng: random.Random,
    out_indexes_used: list,
) -> Generator[Union[Cut, Tuple[Cut, ...]], None, None]:
    """
    Generator which will yield items in a sequence in a random order.
    It will append the indexes of items yielded during iteration via ``out_used_indexes``.
    """
    with bucket.mutex:
        bucket = list(bucket.queue)
    indexes = list(range(len(bucket)))
    rng.shuffle(indexes)
    for idx in indexes:
        out_indexes_used.append(idx)
        yield bucket[idx]


class BucketsDontHaveEnoughData(Exception):
    pass


def _emit_shuffle_buffer_size_warning():
    warnings.warn(
        "Since Lhotse v1.20 'shuffle_buffer_size' is deprecated, because DynamicBucketingSampler "
        "does not require a separate shuffling buffer anymore. "
        "To improve both shuffling and sampling randomness, increase 'buffer_size' instead. "
        "To maintain backward compatibility, we will increase 'buffer_size' "
        "by 'shuffling_buffer_size' for you. "
        "This argument will be deprecated in a future Lhotse version.",
        category=DeprecationWarning,
    )



import logging
from typing import List

import click

from lhotse.bin.modes import download, prepare
from lhotse.recipes.reazonspeech import (
    REAZONSPEECH,
    download_reazonspeech,
    prepare_reazonspeech,
)
from lhotse.utils import Pathlike

__all__ = ["reazonspeech"]


@prepare.command(context_settings=dict(show_default=True))
@click.argument("corpus_dir", type=click.Path(exists=True, dir_okay=True))
@click.argument("output_dir", type=click.Path())
@click.option(
    "-j",
    "--num-jobs",
    type=int,
    default=1,
    help="How many threads to use (can give good speed-ups with slow disks).",
)
def reazonspeech(
    corpus_dir: Pathlike,
    output_dir: Pathlike,
    num_jobs: int,
):
    """ReazonSpeech ASR data preparation."""
    logging.basicConfig(level=logging.INFO)
    prepare_reazonspeech(corpus_dir, output_dir=output_dir, num_jobs=num_jobs)


@download.command(context_settings=dict(show_default=True))
@click.argument("target_dir", type=click.Path())
@click.option(
    "--subset",
    type=click.Choice(("auto",) + REAZONSPEECH),
    multiple=True,
    default=["auto"],
    help="List of dataset parts to prepare (default: small-v1). To prepare multiple parts, pass each with `--subset` "
    "Example: `--subset all",
)
@click.option(
    "-j",
    "--num-jobs",
    type=int,
    default=1,
    help="How many threads to use (can give good speed-ups with slow disks).",
)
def reazonspeech(target_dir: Pathlike, subset: List[str], num_jobs: int):
    """ReazonSpeech download."""
    logging.basicConfig(level=logging.INFO)
    if "auto" in subset:
        subset = "auto"
    download_reazonspeech(target_dir, dataset_parts=subset, num_jobs=num_jobs)



"""
This recipe prepares data collected from radio streamed on the web. The data
have some metadata attached to them, including the geographic location of
broadcast, date and time of the recorded clip, as well as a unique station
identifier.

Obtaining the data
-----------------------------------------------------------
If you want to use this corpus please email: wiesner@jhu.edu

As the data are collected from radio stream, they cannot be broadly
disseminated or used for commercial purposes. In the email, include your
affiliated academic institution and the intended use for the data and we will
the data to you if it is indeed for non-commercial, academic purporses.

Description
------------------------------------------------------------
The data consist of ∼4000 hours of speech collected between
September 27, 2023 to October 1, 2023, in 9449 locations all over the world,
from 17171 stations. 

These data were used for Geolocation of speech in order to answer the question,
Where are you from? in the paper 

Where are you from? Geolocating Speech and Applications to Language
Identification, presented at NAACL 2024. Please read for a full descrption
and please cite as 
 
@inproceedings{foley2024you,
  title={Where are you from? Geolocating Speech and Applications to Language Identification},
  author={Foley, Patrick and Wiesner, Matthew and Odoom, Bismarck and Perera, Leibny Paola Garcia and Murray, Kenton and Koehn, Philipp},
  booktitle={Proceedings of the 2024 Conference of the North American Chapter of the Association for Computational Linguistics: Human Language Technologies (Volume 1: Long Papers)},
  pages={5114--5126},
  year={2024}
}
"""
import json
import re
from functools import partial
from pathlib import Path
from typing import Dict, Optional, Union

from tqdm import tqdm

from lhotse.audio import Recording, RecordingSet
from lhotse.parallel import parallel_map
from lhotse.supervision import SupervisionSegment, SupervisionSet
from lhotse.utils import Pathlike


def _make_reco_and_sups_from_file(sf: str, msd: float = 0.5):
    corpus_dir = sf.parents[2]
    audio_dir = corpus_dir / "recos"
    fname = sf.with_suffix(".flac").stem

    # E.g. 2023_10_01_09h_02m_54s_dur30_ZnpbY9Zx_lat3.17_long113.04
    chunk_idx = int(sf.parent.suffix.strip("."))
    reco_file = audio_dir / f"recos.{chunk_idx}" / f"{fname}.flac"
    reco = Recording.from_file(reco_file, recording_id=fname)
    reco.channel_ids = [0]
    sups = []
    total = 0
    with open(sf) as f:
        segments = json.load(f)

    # Parse the file format, shown in the comment above, to get:
    # date, station, latitude, longitude, and the estimated gender
    lat, lon = re.search(r"lat[^_]+_long[^_]+", Path(sf).stem).group(0).split("_")
    lat = float(lat.replace("lat", ""))
    lon = float(lon.replace("long", ""))
    station = re.search(r"s_dur[0-9]+_(.*)_lat[^_]+_long[^_]+", fname).groups()[0]
    fname_vals = fname.split("_")
    date = [int(i.strip("hms")) for i in fname_vals[0:6]]  # YY MM DD hh mm ss
    for seg in segments:
        start, end = float(seg[1]), float(seg[2])
        dur = end - start
        if seg[0] in ("male", "female") and dur > msd:
            sups.append(
                SupervisionSegment(
                    id=f"{fname}_{int(100*start):04}",
                    recording_id=fname,
                    start=start,
                    duration=round(dur, 4),
                    channel=0,
                    custom={
                        "date": date,
                        "lat": lat,
                        "lon": lon,
                        "station": station,
                        "est_gender": seg[0],
                    },
                )
            )
    return sups, reco


def prepare_radio(
    corpus_dir: Pathlike,
    output_dir: Optional[Pathlike] = None,
    min_segment_duration: float = 0.5,
    num_jobs: int = 4,
) -> Dict[str, Union[RecordingSet, SupervisionSet]]:
    """
    Return the manifests which consist of recordings and supervisions
    :param corpus_dir: Path to the collected radio samples
    :param output_dir: Pathlike, the path where manifests are written
    :return: A Dict whose key is the dataset part and the value is a Dict with
        keys 'recordings' and 'supervisions'.
    """
    corpus_dir = Path(corpus_dir)
    segment_files = corpus_dir.rglob("segs/*/*.json")
    supervisions, recordings = [], []
    fun = partial(_make_reco_and_sups_from_file, msd=min_segment_duration)
    output_dir = Path(output_dir) if output_dir is not None else None
    output_dir.mkdir(mode=511, parents=True, exist_ok=True)
    with RecordingSet.open_writer(
        output_dir / "radio_recordings.jsonl.gz"
    ) as rec_writer:
        with SupervisionSet.open_writer(
            output_dir / "radio_supervisions.jsonl.gz"
        ) as sup_writer:
            for sups, reco in tqdm(
                parallel_map(
                    fun,
                    segment_files,
                    num_jobs=num_jobs,
                ),
                desc=f"Making recordings and supervisions",
            ):
                rec_writer.write(reco)
                for sup in sups:
                    sup_writer.write(sup)

            manifests = {
                "recordings": RecordingSet.from_jsonl_lazy(rec_writer.path),
                "supervisions": SupervisionSet.from_jsonl_lazy(sup_writer.path),
            }

    return manifests



import logging
from typing import Generator, List, Optional, Union

import torch

from lhotse import CutSet, MonoCut, RecordingSet, SupervisionSegment, add_durations
from lhotse.qa import trim_supervisions_to_recordings
from lhotse.utils import fastcopy, is_module_available


def annotate_with_whisper(
    manifest: Union[RecordingSet, CutSet],
    model_name: str = "base",
    device: str = "cpu",
    force_nonoverlapping: bool = False,
    download_root: Optional[str] = None,
    **decode_options,
) -> Generator[MonoCut, None, None]:
    """
    Use OpenAI Whisper model to annotate either RECORDINGS_MANIFEST, RECORDINGS_DIR, or CUTS_MANIFEST.
    It will perform automatic segmentation, transcription, and language identification. If
    the first argument is a CutSet, it will overwrite the supervisions with the results of the inference.

    Note: this is an experimental feature of Lhotse, and is not guaranteed to yield
    high quality of data.

    See the original repo for more details: https://github.com/openai/whisper

    :param manifest: a ``RecordingSet`` or ``CutSet`` object.
    :param language: specify the language if known upfront, otherwise it will be auto-detected.
    :param model_name: one of available Whisper variants (base, medium, large, etc.).
    :param device: Where to run the inference (cpu, cuda, etc.).
    :param force_nonoverlapping: if True, the Whisper segment time-stamps will be processed to make
        sure they are non-overlapping.
    :param download_root: if specified, the model will be downloaded to this directory. Otherwise,
        it will be downloaded to the default location specfied by whisper.
    :param decode_options: additional options to pass to the ``whisper.transcribe`` function.
    :return: a generator of cuts (use ``CutSet.open_writer()`` to write them).
    """
    assert is_module_available("whisper"), (
        "This function expects OpenAI Whisper to be installed. "
        "You can install it via 'pip install git+https://github.com/openai/whisper.git' "
        "(see https://github.com/openai/whisper for details)."
    )

    if isinstance(manifest, RecordingSet):
        yield from _annotate_recordings(
            manifest,
            model_name,
            device,
            force_nonoverlapping,
            download_root,
            **decode_options,
        )
    elif isinstance(manifest, CutSet):
        yield from _annotate_cuts(
            manifest,
            model_name,
            device,
            force_nonoverlapping,
            download_root,
            **decode_options,
        )
    else:
        raise ValueError("The ``manifest`` must be either a RecordingSet or a CutSet.")


def _annotate_recordings(
    recordings: RecordingSet,
    model_name: str,
    device: str,
    force_nonoverlapping: bool,
    download_root: Optional[str] = None,
    **decode_options,
):
    """
    Helper function that annotates a RecordingSet with Whisper.
    """
    import whisper

    model = whisper.load_model(model_name, device=device, download_root=download_root)

    for recording in recordings:
        if recording.num_channels > 1:
            logging.warning(
                f"Skipping recording '{recording.id}'. It has {recording.num_channels} channels, "
                f"but we currently only support mono input."
            )
            continue
        audio = torch.from_numpy(recording.resample(16000).load_audio()).squeeze(0)
        result = whisper.transcribe(model=model, audio=audio, **decode_options)
        # Create supervisions from segments while filtering out those with negative duration.
        supervisions = [
            SupervisionSegment(
                id=f"{recording.id}-{segment['id']:06d}",
                recording_id=recording.id,
                start=round(segment["start"], ndigits=8),
                duration=add_durations(
                    segment["end"], -segment["start"], sampling_rate=16000
                ),
                text=segment["text"].strip(),
                language=result["language"],
            )
            for segment in result["segments"]
            if segment["end"] - segment["start"] > 0
        ]
        cut = recording.to_cut()
        if supervisions:
            supervisions = (
                _postprocess_timestamps(supervisions)
                if force_nonoverlapping
                else supervisions
            )
            cut.supervisions = list(
                trim_supervisions_to_recordings(
                    recordings=recording, supervisions=supervisions, verbose=False
                )
            )
        yield cut


def _annotate_cuts(
    cuts: CutSet,
    model_name: str,
    device: str,
    force_nonoverlapping: bool,
    download_root: Optional[str] = None,
    **decode_options,
):
    """
    Helper function that annotates a CutSet with Whisper.
    """
    import whisper

    model = whisper.load_model(model_name, device=device, download_root=download_root)

    for cut in cuts:
        if cut.num_channels > 1:
            logging.warning(
                f"Skipping cut '{cut.id}'. It has {cut.num_channels} channels, "
                f"but we currently only support mono input."
            )
            continue
        audio = torch.from_numpy(cut.resample(16000).load_audio()).squeeze(0)
        result = whisper.transcribe(model=model, audio=audio, **decode_options)
        # Create supervisions from segments while filtering out those with negative duration.
        supervisions = [
            SupervisionSegment(
                id=f"{cut.id}-{segment['id']:06d}",
                recording_id=cut.recording_id,
                start=round(segment["start"], ndigits=8),
                duration=add_durations(
                    min(segment["end"], cut.duration),
                    -segment["start"],
                    sampling_rate=16000,
                ),
                text=segment["text"].strip(),
                language=result["language"],
            )
            for segment in result["segments"]
            if segment["end"] - segment["start"] > 0
        ]
        new_cut = fastcopy(
            cut,
            supervisions=_postprocess_timestamps(supervisions)
            if force_nonoverlapping
            else supervisions,
        )
        yield new_cut


def _postprocess_timestamps(supervisions: List[SupervisionSegment]):
    """
    Whisper tends to have a lot of overlapping segments due to inaccurate end timestamps.
    Under a strong assumption that the input speech is non-overlapping, we can fix that
    by always truncating to the start timestamp of the next segment.
    """
    from cytoolz import sliding_window

    supervisions = sorted(supervisions, key=lambda s: s.start)

    if len(supervisions) < 2:
        return supervisions
    out = []
    for cur, nxt in sliding_window(2, supervisions):
        if cur.end > nxt.start:
            cur = cur.trim(end=nxt.start)
        out.append(cur)
    out.append(nxt)
    return out



"""
Silero VAD integration module

This module integrates Silero VAD (Voice Activity Detector),
a high-precision pre-trained model for voice activity detection in audio streams.

GitHub Repository: https://github.com/snakers4/silero-vad

Citations:

@misc{Silero VAD,
  author = {Silero Team},
  title = {
      Silero VAD: pre-trained enterprise-grade Voice Activity Detector (VAD),
      Number Detector and Language Classifier
  },
  year = {2021},
  publisher = {GitHub},
  journal = {GitHub repository},
  howpublished = {\\url{https://github.com/snakers4/silero-vad}},
  commit = {insert_some_commit_here},
  email = {hello@silero.ai}
}

Please cite Silero VAD when used in your projects. Details can be found in the repository.
"""

import shutil
from contextlib import suppress
from pathlib import Path
from typing import Dict, List

import numpy as np
import torch

from .base import Activity, ActivityDetector


def _to_activity_maker(sampling_rate: int):
    def make_activity(state: Dict[str, float]) -> Activity:
        start, end = state["start"], state["end"]
        return Activity(
            start=start / sampling_rate,
            duration=(end - start) / sampling_rate,
        )

    return make_activity


class SileroVAD(ActivityDetector):
    """Silero Voice Activity Detector model wrapper"""

    def __init__(
        self,
        *,
        device: str = "cpu",
        sampling_rate: int = 16_000,
        force_download: bool = False,
    ):
        if sampling_rate not in [8_000, 16_000]:  # pragma: no cover
            msg = (
                "Sampling rate must be either 8000 or 16000, ",
                f"but got {sampling_rate}",
            )
            raise ValueError(msg)
        super().__init__(
            detector_name=f"silero_vad_{sampling_rate//1000}k",
            device=device,
            sampling_rate=sampling_rate,
        )

        self._model, utils = self._get_model(force_download=force_download)

        # utils[0] := get_speech_timestamps - function that returns speech timestamps
        self._predict = utils[0]
        self._model.to(self.device)
        self._to_activity = _to_activity_maker(sampling_rate)

    @classmethod
    def _cache_dirs(cls) -> List[Path]:
        cache_dir = torch.hub.get_dir()  # type: ignore
        if not isinstance(cache_dir, str):  # pragma: no cover
            raise TypeError(f"Bad cache directory path. Got {cache_dir}")
        return list(Path(cache_dir).glob("snakers4_silero-vad_*"))

    @classmethod
    def _clear_cache(cls):  # pragma: no cover
        """Remove Silero VAD models from cache"""
        for directory in cls._cache_dirs():
            if directory.is_dir():
                shutil.rmtree(directory)

    @classmethod
    def _get_model(cls, *, force_download: bool = False):
        if force_download:  # pragma: no cover
            cls._clear_cache()
        if not cls._cache_dirs():  # pragma: no cover
            force_download = True

        config = {
            "repo_or_dir": "snakers4/silero-vad",
            "model": "silero_vad",
            "onnx": False,
            "force_reload": force_download,
        }
        with suppress(Exception):
            from pkg_resources import parse_version  # pylint: disable=C0415

            # trust_repo is a new parameter in torch.hub.load and requires torch >= 2.0
            if parse_version(torch.__version__) >= parse_version("1.12"):
                config["trust_repo"] = True
        return torch.hub.load(**config)  # type: ignore

    def forward(self, track: np.ndarray) -> List[Activity]:
        """Predict voice activity for audio"""
        audio = torch.Tensor(track).to(self.device)

        with torch.no_grad():
            murkup = self._predict(
                audio=audio,
                model=self._model,
                sampling_rate=self._sampling_rate,
                min_speech_duration_ms=250,
                max_speech_duration_s=float("inf"),
                min_silence_duration_ms=100,
                window_size_samples=512,
                speech_pad_ms=30,
                return_seconds=False,
                visualize_probs=False,
                progress_tracking_callback=None,
            )

        return list(map(self._to_activity, murkup))

    @classmethod
    def force_download(cls):  # pragma: no cover
        cls._get_model(force_download=True)


class SileroVAD8k(SileroVAD):
    def __init__(self, device: str = "cpu", *, force_download: bool = False):
        super().__init__(
            sampling_rate=8_000,
            device=device,
            force_download=force_download,
        )


class SileroVAD16k(SileroVAD):
    def __init__(self, device: str = "cpu", *, force_download: bool = False):
        super().__init__(
            sampling_rate=16_000,
            device=device,
            force_download=force_download,
        )



"""
About the CDSD (Chinese Dysarthric Speech Database) dataset:

    This database comprises speech data from 24 participants with dysarthria. 
    
    Among these participants, one recorded an additional 10 hours of speech data, while each recorded one hour, resulting in 34 hours of speech material. 

    To accommodate participants with varying cognitive levels, the text pool primarily consists of content from the AISHELL-1 dataset and speeches by primary and secondary school students. When participants read these texts, they must use a mobile device or the ZOOM F8n multi-track field recorder to record their speeches. 
    
    In this paper, the authors elucidate the data collection and annotation processes and present an approach for establishing a baseline for dysarthric speech recognition. Furthermore, the authors conducted a speaker-dependent dysarthric speech recognition experiment using an additional 10 hours of speech data from one of the participants. 
    
arXiv link: https://arxiv.org/abs/2310.15930v1
"""

import logging
from collections import defaultdict
from pathlib import Path
from typing import Dict, Optional, Union

from tqdm.auto import tqdm

from lhotse import validate_recordings_and_supervisions
from lhotse.audio import Recording, RecordingSet
from lhotse.qa import fix_manifests
from lhotse.supervision import SupervisionSegment, SupervisionSet
from lhotse.utils import Pathlike


def text_normalize(line: str):
    """
    Modified from https://github.com/wenet-e2e/wenet/blob/main/examples/multi_cn/s0/local/aishell_data_prep.sh#L54
    sed 's/ａ/a/g' | sed 's/ｂ/b/g' |\
    sed 's/ｃ/c/g' | sed 's/ｋ/k/g' |\
    sed 's/ｔ/t/g' > $dir/transcripts.t

    """
    line = line.replace("ａ", "a")
    line = line.replace("ｂ", "b")
    line = line.replace("ｃ", "c")
    line = line.replace("ｋ", "k")
    line = line.replace("ｔ", "t")
    line = line.upper()
    return line


def prepare_cdsd(
    corpus_dir: Pathlike, output_dir: Optional[Pathlike] = None
) -> Dict[str, Dict[str, Union[RecordingSet, SupervisionSet]]]:
    """
    Returns the manifests which consist of the Recordings and Supervisions
    :param corpus_dir: Pathlike, the path of the data dir.
    :param output_dir: Pathlike, the path where to write the manifests.
    :return: a Dict whose key is the dataset part, and the value is Dicts with the keys 'recordings' and 'supervisions'.
    """
    corpus_dir = Path(corpus_dir)
    assert corpus_dir.is_dir(), f"No such directory: {corpus_dir}"
    if output_dir is not None:
        output_dir = Path(output_dir)
        output_dir.mkdir(parents=True, exist_ok=True)

    manifests = defaultdict(dict)
    dataset_parts = ["1h", "10h"]
    for part in dataset_parts:
        logging.info(f"Processing CDSD subset: {part}")
        recordings = []
        supervisions = []

        txt_path = corpus_dir / "after_catting" / f"{part}" / "Text"
        transcript_dict = {}
        for text_path in txt_path.rglob("**/*.txt"):
            with open(text_path, "r", encoding="utf-8") as f:
                for line in f.readlines():
                    idx_transcript, content = line.strip().split(maxsplit=1)
                    content = text_normalize(content)
                    transcript_dict[idx_transcript] = content

        wav_path = corpus_dir / "after_catting" / f"{part}" / "Audio"
        for audio_path in tqdm(wav_path.rglob("**/*.wav"), desc="Processing audio"):
            idx = audio_path.stem
            speaker = audio_path.parts[-2]
            if idx not in transcript_dict:
                logging.warning(f"No transcript: {idx}")
                logging.warning(f"{audio_path} has no transcript.")
                continue
            text = transcript_dict[idx]
            if not audio_path.is_file():
                logging.warning(f"No such file: {audio_path}")
                continue
            recording = Recording.from_file(audio_path)
            recordings.append(recording)
            segment = SupervisionSegment(
                id=idx,
                recording_id=idx,
                start=0.0,
                duration=recording.duration,
                channel=0,
                language="Chinese",
                speaker=speaker,
                text=text.strip().replace(" ", ""),
                # here we remove the space between words in the text
                # in advance.
            )
            supervisions.append(segment)

        recording_set = RecordingSet.from_recordings(recordings)
        supervision_set = SupervisionSet.from_segments(supervisions)
        recording_set, supervision_set = fix_manifests(recording_set, supervision_set)
        validate_recordings_and_supervisions(recording_set, supervision_set)

        if output_dir is not None:
            supervision_set.to_file(output_dir / f"cdsd_supervisions_{part}.jsonl.gz")
            recording_set.to_file(output_dir / f"cdsd_recordings_{part}.jsonl.gz")

        manifests[part] = {"recordings": recording_set, "supervisions": supervision_set}

    return manifests



"""
ReazonSpeech is an open-source dataset that contains a diverse set of natural Japanese speech,
collected from terrestrial television streams. It contains more than 35,000 hours of audio.

The dataset is available on Hugging Face. For more details, please visit:

Dataset: https://huggingface.co/datasets/reazon-research/reazonspeech
Paper: https://research.reazon.jp/_static/reazonspeech_nlp2023.pdf
"""

import json
import logging
import re
from collections import defaultdict
from pathlib import Path
from typing import Any, Dict, Optional, Sequence, Tuple, Union

from tqdm.auto import tqdm

from lhotse import CutSet, fix_manifests, validate_recordings_and_supervisions
from lhotse.audio import Recording, RecordingSet
from lhotse.parallel import parallel_map
from lhotse.recipes.utils import manifests_exist, read_manifests_if_cached
from lhotse.supervision import SupervisionSegment, SupervisionSet
from lhotse.utils import Pathlike, is_module_available

REAZONSPEECH = (
    "tiny",
    "small",
    "medium",
    "large",
    "all",
    "small-v1",
    "medium-v1",
    "all-v1",
)

PUNCTUATIONS = {ord(x): "" for x in "、。「」『』，,？！!!?!?"}
ZENKAKU = "ａｂｃｄｅｆｇｈｉｊｋｌｍｎｏｐｑｒｓｔｕｖｗｘｙｚＡＢＣＤＥＦＧＨＩＪＫＬＭＮＯＰＱＲＳＴＵＶＷＸＹＺ０１２３４５６７８９"
HANKAKU = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"
ZEN2HAN = str.maketrans(ZENKAKU, HANKAKU)


def normalize(s):
    """
    Convert full-width characters to half-width, and remove punctuations.
    :param s: str, input string.
    :return: str, normalized string.
    """
    if is_module_available("num2words"):
        import num2words
    else:
        raise ImportError(
            "To process the ReazonSpeech corpus, please install optional dependency: pip install num2words"
        )
    s = s.translate(PUNCTUATIONS).translate(ZEN2HAN)
    conv = lambda m: num2words.num2words(m.group(0), lang="ja")
    return re.sub(r"\d+\.?\d*", conv, s)


def write_to_json(data, filename):
    """
    Writes data to a JSON file.
    :param data: The data to write.
    :param filename: The name of the file to write to.
    """

    with open(filename, "w", encoding="utf-8") as f:
        json.dump(data, f, ensure_ascii=False, indent=4)


def download_reazonspeech(
    target_dir: Pathlike = ".",
    dataset_parts: Optional[Union[str, Sequence[str]]] = "auto",
    num_jobs: int = 1,
) -> Path:
    """
    Download the ReazonSpeech dataset.
    :param target_dir: Pathlike, the path of the dir to storage the dataset.
    :param dataset_parts: the parts of the dataset to download (e.g. small, medium, or large).
    :param num_jobs: the number of processes to download and format.
    :return: the path to downloaded data and the JSON file.
    """
    if is_module_available("datasets"):
        import soundfile as sf
        from datasets import Audio, load_dataset
    else:
        raise ImportError(
            "To process the ReazonSpeech corpus, please install optional dependencies: pip install datasets soundfile"
        )
    target_dir = Path(target_dir)
    target_dir.mkdir(parents=True, exist_ok=True)
    corpus_dir = target_dir / "ReazonSpeech"

    if dataset_parts == "auto":
        dataset_parts = ("small-v1",)
    elif isinstance(dataset_parts, str):
        dataset_parts = [dataset_parts]

    for part in dataset_parts:
        logging.info(f"Downloading ReazonSpeech part: {part}")
        ds = load_dataset(
            "reazon-research/reazonspeech",
            part,
            trust_remote_code=True,
            cache_dir=corpus_dir,
            num_proc=num_jobs,
        )["train"]

    # Prepare data for JSON export
    def format_example(example: dict, idx: int) -> dict:
        example["id"] = str(idx)
        example["audio_filepath"] = example["audio"]["path"]
        example["text"] = normalize(example["transcription"])
        example["duration"] = sf.info(example["audio"]["path"]).duration
        return example

    ds = ds.cast_column("audio", Audio(decode=True))  # Hack: don't decode to speedup
    ds = ds.map(
        format_example,
        with_indices=True,
        remove_columns=ds.column_names,
        num_proc=num_jobs,
    )

    # Write data to a JSON file
    ds.to_json(
        corpus_dir / "dataset.json",
        num_proc=num_jobs,
        force_ascii=False,
        indent=4,
        lines=False,
        batch_size=ds.num_rows,
    )

    return corpus_dir


def prepare_reazonspeech(
    corpus_dir: Pathlike,
    output_dir: Optional[Pathlike],
    num_jobs: int = 1,
) -> Dict[str, Dict[str, Union[RecordingSet, SupervisionSet]]]:
    """
    Returns the manifests which consist of the Recordings and Supervisions.
    When all the manifests are available in the ``output_dir``, it will simply read and return them.
    :param corpus_dir: Pathlike, the path of the data dir.
    :param output_dir: Pathlike, the path where to write the manifests.
    :param num_jobs: int, number of parallel threads used for 'parse_utterance' calls.
    :return: a Dict whose key is the dataset part, and the value is Dicts with the keys 'audio' and 'supervisions'.
    """
    corpus_dir = Path(corpus_dir)
    assert corpus_dir.is_dir(), f"No such directory: {corpus_dir}"

    # Split the dataset into train, dev, and test
    with open(corpus_dir / "dataset.json", "r", encoding="utf-8") as file:
        full = json.load(file)
        dev = full[:1000]
        test = full[1000:1100]
        train = full[1100:]

        write_to_json(train, corpus_dir / "train.json")
        write_to_json(dev, corpus_dir / "dev.json")
        write_to_json(test, corpus_dir / "test.json")

    parts = ("train", "dev", "test")

    output_dir = Path(output_dir)
    output_dir.mkdir(parents=True, exist_ok=True)
    # Maybe some manifests already exist: we can read them and save a bit of preparation time.
    manifests = read_manifests_if_cached(
        dataset_parts=parts,
        output_dir=output_dir,
        prefix="reazonspeech",
        suffix="jsonl.gz",
        lazy=True,
    )

    for part in parts:
        logging.info(f"Processing ReazonSpeech subset: {part}")
        if manifests_exist(
            part=part, output_dir=output_dir, prefix="reazonspeech", suffix="jsonl.gz"
        ):
            logging.info(f"ReazonSpeech subset: {part} already prepared - skipping.")
            continue

        filename = corpus_dir / f"{part}.json"
        with open(filename, "r", encoding="utf-8") as file:
            items = json.load(file)

        with RecordingSet.open_writer(
            output_dir / f"reazonspeech_recordings_{part}.jsonl.gz"
        ) as rec_writer, SupervisionSet.open_writer(
            output_dir / f"reazonspeech_supervisions_{part}.jsonl.gz"
        ) as sup_writer, CutSet.open_writer(
            output_dir / f"reazonspeech_cuts_{part}.jsonl.gz"
        ) as cut_writer:
            for recording, segment in tqdm(
                parallel_map(
                    parse_utterance,
                    items,
                    num_jobs=num_jobs,
                ),
                desc="Processing reazonspeech JSON entries",
            ):
                # Fix and validate the recording + supervisions
                recordings, segments = fix_manifests(
                    recordings=RecordingSet.from_recordings([recording]),
                    supervisions=SupervisionSet.from_segments([segment]),
                )
                validate_recordings_and_supervisions(
                    recordings=recordings, supervisions=segments
                )
                # Create the cut since most users will need it anyway.
                # There will be exactly one cut since there's exactly one recording.
                cuts = CutSet.from_manifests(
                    recordings=recordings, supervisions=segments
                )
                # Write the manifests
                rec_writer.write(recordings[0])
                sup_writer.write(segments[0])
                cut_writer.write(cuts[0])

        manifests[part] = {
            "recordings": RecordingSet.from_jsonl_lazy(rec_writer.path),
            "supervisions": SupervisionSet.from_jsonl_lazy(sup_writer.path),
            "cuts": CutSet.from_jsonl_lazy(cut_writer.path),
        }

    return dict(manifests)


def parse_utterance(item: Any) -> Optional[Tuple[Recording, SupervisionSegment]]:
    """
    Process a single utterance from the ReazonSpeech dataset.
    :param item: The utterance to process.
    :return: A tuple containing the Recording and SupervisionSegment.
    """
    recording = Recording.from_file(item["audio_filepath"], recording_id=item["id"])
    segments = SupervisionSegment(
        id=item["id"],
        recording_id=item["id"],
        start=0.0,
        duration=item["duration"],
        channel=0,
        language="Japanese",
        text=item["text"],
    )
    return recording, segments



import logging
from typing import List

import click

from lhotse.bin.modes import download, prepare
from lhotse.recipes.gigast import (
    GIGASPEECH_PARTS,
    GIGAST_LANGS,
    download_gigast,
    prepare_gigast,
)
from lhotse.utils import Pathlike


@prepare.command(context_settings=dict(show_default=True))
@click.argument("corpus_dir", type=click.Path(exists=True, dir_okay=True))
@click.argument("manifests_dir", type=click.Path(exists=True, dir_okay=True))
@click.argument("output_dir", type=click.Path())
@click.option(
    "-l",
    "--language",
    type=click.Choice(("auto",) + GIGAST_LANGS),
    default=["auto"],
    multiple=True,
    help="Languages to download. one of: 'all' (downloads all known languages); a single language code (e.g., 'en')",
)
@click.option(
    "--subset",
    type=click.Choice(("auto",) + GIGASPEECH_PARTS),
    multiple=True,
    default=["auto"],
    help="Which parts of Gigaspeech to download (by default XL + DEV + TEST).",
)
def gigast(
    corpus_dir: Pathlike,
    manifests_dir: Pathlike,
    output_dir: Pathlike,
    language: List[str],
    subset: List[str],
):
    """GigaST data preparation."""
    languages = language[0] if len(language) == 1 else language
    if "auto" in subset:
        subset = "auto"
    prepare_gigast(
        corpus_dir=corpus_dir,
        manifests_dir=manifests_dir,
        output_dir=output_dir,
        languages=languages,
        dataset_parts=subset,
    )


@download.command(context_settings=dict(show_default=True))
@click.argument("target_dir", type=click.Path())
@click.option(
    "-l",
    "--languages",
    default="all",
    help="Languages to download. one of: 'all' (downloads all known languages); a single language code (e.g., 'en')",
)
@click.option("--force-download", is_flag=True, default=False, help="Force download")
def gigast(
    target_dir: Pathlike,
    languages: List[str],
    force_download: bool = False,
):
    """GigaST download."""
    download_gigast(
        target_dir=target_dir,
        languages=languages,
        force_download=force_download,
    )



from typing import Optional, Sequence

import click

from lhotse.bin.modes import download, prepare
from lhotse.recipes.mdcc import download_mdcc, prepare_mdcc
from lhotse.utils import Pathlike


@prepare.command(context_settings=dict(show_default=True))
@click.argument("corpus_dir", type=click.Path(exists=True, dir_okay=True))
@click.argument("output_dir", type=click.Path())
@click.option(
    "-p",
    "--dataset-parts",
    type=str,
    default=["all"],
    multiple=True,
    help="List of dataset parts to prepare. To prepare multiple parts, pass each with `-p` "
    "Example: `-p train -p valid`",
)
def MDCC(
    corpus_dir: Pathlike,
    dataset_parts: Sequence[str],
    output_dir: Optional[Pathlike] = None,
):
    """MDCC data preparation."""
    prepare_mdcc(
        corpus_dir=corpus_dir,
        dataset_parts=dataset_parts,
        output_dir=output_dir,
    )


@download.command(context_settings=dict(show_default=True))
@click.argument("target_dir", type=click.Path())
@click.option(
    "--force-download",
    is_flag=True,
    default=False,
    help="if True, it will download the MDCC data even if it is already present.",
)
def MDCC(
    target_dir: Pathlike,
    force_download: Optional[bool] = False,
):
    """MDCC download."""
    download_mdcc(
        target_dir=target_dir,
        force_download=force_download,
    )



from typing import Dict, List, Optional, Tuple, Union

import click

from lhotse.bin.modes import download, prepare
from lhotse.recipes.medical import download_medical, prepare_medical
from lhotse.utils import Pathlike


@prepare.command(context_settings=dict(show_default=True))
@click.argument("corpus_dir", type=click.Path(exists=True, dir_okay=True))
@click.argument("output_dir", type=click.Path())
@click.option(
    "-j",
    "--num-jobs",
    type=int,
    default=1,
    help="How many threads to use (can give good speed-ups with slow disks).",
)
def medical(
    corpus_dir: Pathlike,
    output_dir: Optional[Pathlike] = None,
    num_jobs: int = 1,
):
    """Medical data preparation."""
    prepare_medical(
        corpus_dir=corpus_dir,
        output_dir=output_dir,
        num_jobs=num_jobs,
    )


@download.command(context_settings=dict(show_default=True))
@click.argument("target_dir", type=click.Path())
@click.option("--force-download", is_flag=True, default=False, help="Force download")
def medical(
    target_dir: Pathlike,
    force_download: Optional[bool] = False,
):
    """Medical download."""
    download_medical(
        target_dir=target_dir,
        force_download=force_download,
    )



"""
About the speech_commands corpus
This is a set of one-second .wav audio files, each containing a single spoken English word or background noise.
These words are from a small set of commands, and are spoken by a variety of different speakers.
This data set is designed to help train simple machine learning models.
It is covered in more detail at https://arxiv.org/abs/1804.03209.

Version 0.01 of the data set (configuration "v0.01") was released on August 3rd 2017 and contains 64,727 audio files.
Version 0.02 of the data set (configuration "v0.02") was released on April 11th 2018 and contains 105,829 audio files.
"""

import logging
import os
import shutil
import tarfile
from collections import defaultdict
from pathlib import Path
from typing import Dict, List, Optional, Sequence, Tuple, Union

from tqdm.auto import tqdm

from lhotse import fix_manifests, validate_recordings_and_supervisions
from lhotse.audio import Recording, RecordingSet
from lhotse.recipes.utils import manifests_exist, read_manifests_if_cached
from lhotse.supervision import SupervisionSegment, SupervisionSet
from lhotse.utils import Pathlike, is_module_available, resumable_download, safe_extract

_DOWNLOAD_PATH_V1 = "http://download.tensorflow.org/data/speech_commands_v0.01.tar.gz"
_TEST_DOWNLOAD_PATH_V1 = (
    "http://download.tensorflow.org/data/speech_commands_test_set_v0.01.tar.gz"
)
_DOWNLOAD_PATH_V2 = "http://download.tensorflow.org/data/speech_commands_v0.02.tar.gz"
_TEST_DOWNLOAD_PATH_V2 = (
    "http://download.tensorflow.org/data/speech_commands_test_set_v0.02.tar.gz"
)

_SPLITS = ["train", "valid", "test"]

WORDS = ["down", "go", "left", "no", "off", "on", "right", "stop", "up", "yes"]
SILENCE = "_silence_"
UNKNOWN = "_unknown_"
BACKGROUND_NOISE = "_background_noise_"


def download_speechcommands(
    speechcommands_version: str = "2",
    target_dir: Pathlike = ".",
    force_download: bool = False,
) -> Path:
    """
    Download and unzip Speech Commands dataset

    :param speechcommands_version: str, dataset version.
    :param target_dir: Pathlike, the path of the dir to store the dataset.
    :param force_download: bool, if True, download the archive even if it already exists.

    :return: the path to downloaded and extracted directory with data.
    """
    target_dir = Path(target_dir)
    corpus_dir = target_dir / f"SpeechCommands{speechcommands_version}"
    corpus_dir.mkdir(parents=True, exist_ok=True)

    dataset_parts = (
        f"speech_commands_v0.0{speechcommands_version}",
        f"speech_commands_test_set_v0.0{speechcommands_version}",
    )

    for part in tqdm(
        dataset_parts, desc=f"Downloading Speech Commands v0.0{speechcommands_version}"
    ):
        logging.info(f"Processing split: {part}")
        part_dir = corpus_dir / part
        completed_detector = part_dir / ".completed"
        if completed_detector.is_file():
            logging.info(f"Skipping {part} because {completed_detector} exists.")
            continue
        # Process the archive.
        tar_name = f"{part}.tar.gz"
        tar_path = corpus_dir / tar_name
        resumable_download(
            f"http://download.tensorflow.org/data/{tar_name}",
            filename=tar_path,
            force_download=force_download,
        )
        # Remove partial unpacked files, if any, and unpack everything.
        shutil.rmtree(part_dir, ignore_errors=True)
        with tarfile.open(tar_path) as tar:
            tar.extractall(part_dir)
        completed_detector.touch()

    return corpus_dir


def _split_train_valid(
    speechcommands_version: str,
    corpus_dir: Pathlike,
) -> Tuple[set, set]:
    """
    Returns the train_paths and valid_paths given a dataset part.
    :param speechcommands_version: str, dataset version.
    :param corpus_dir: Pathlike, the path of the data dir.
    :return: the train_paths and valid_paths.
    """
    corpus_dir = Path(corpus_dir)
    part_path = corpus_dir / f"speech_commands_v0.0{speechcommands_version}"

    # Split dataset into train and valid
    train_paths = []
    for root in os.listdir(part_path):
        if root == "testing_list.txt":
            with open(part_path / root) as file_obj:
                train_test_paths = file_obj.read().strip().splitlines()
        elif root == "validation_list.txt":
            with open(part_path / root) as file_obj:
                valid_paths = file_obj.read().strip().splitlines()
        elif os.path.isdir(part_path / root):
            file_paths = os.listdir(part_path / root)
            train_paths += [
                os.path.join(root, file_path)
                for file_path in file_paths
                if file_path.endswith(".wav")
            ]

    # Original validation files don't include silence - we add them manually here
    valid_paths.append(os.path.join(BACKGROUND_NOISE, "running_tap.wav"))

    # The paths for the train set is just whichever paths that do not exist in either the test or validation splits.
    train_paths = set(train_paths) - set(valid_paths) - set(train_test_paths)

    return train_paths, valid_paths


def _prepare_train(
    speechcommands_version: str,
    train_paths: set,
    corpus_dir: Pathlike,
) -> Tuple[RecordingSet, SupervisionSet]:
    """
    Returns the RecodingSet and SupervisionSet given a dataset part.
    :param speechcommands_version: str, dataset version.
    :param train_paths: set, the train subset paths.
    :param corpus_dir: Pathlike, the path of the data dir.
    :return: the RecodingSet and SupervisionSet for train and valid.
    """
    corpus_dir = Path(corpus_dir)
    part_path = corpus_dir / f"speech_commands_v0.0{speechcommands_version}"

    train_recordings = []
    train_supervisions = []
    for train_path in train_paths:
        audio_path = part_path / train_path
        audio_path = audio_path.resolve()
        train_path_splits = train_path.split("/")
        audio_file_name = train_path.replace("/", "_").replace(".wav", "")
        audio_file_name_splits = train_path_splits[1].split("_")

        if train_path_splits[0] == BACKGROUND_NOISE:
            speaker = None
            text = ""
        else:
            speaker = audio_file_name_splits[0]
            text = train_path_splits[0].strip()
        if not audio_path.is_file():
            logging.warning(f"No such file: {audio_path}")
            continue
        train_recording = Recording.from_file(
            path=audio_path,
            recording_id=audio_file_name,
        )
        train_recordings.append(train_recording)
        train_segment = SupervisionSegment(
            id=audio_file_name,
            recording_id=audio_file_name,
            start=0.0,
            duration=train_recording.duration,
            channel=0,
            language="English",
            speaker=speaker,
            text=text,
        )
        train_supervisions.append(train_segment)
    train_recording_set = RecordingSet.from_recordings(train_recordings)
    train_supervision_set = SupervisionSet.from_segments(train_supervisions)

    # Fix manifests
    train_recording_set, train_supervision_set = fix_manifests(
        train_recording_set, train_supervision_set
    )
    validate_recordings_and_supervisions(train_recording_set, train_supervision_set)

    return train_recording_set, train_supervision_set


def _prepare_valid(
    speechcommands_version: str,
    valid_paths: set,
    corpus_dir: Pathlike,
) -> Tuple[RecordingSet, SupervisionSet]:
    """
    Returns the RecodingSet and SupervisionSet given a dataset part.
    :param speechcommands_version: str, dataset version.
    :param valid_paths: set, the valid subset paths.
    :param corpus_dir: Pathlike, the path of the data dir.
    :return: the RecodingSet and SupervisionSet for train and valid.
    """
    corpus_dir = Path(corpus_dir)
    part_path = corpus_dir / f"speech_commands_v0.0{speechcommands_version}"

    valid_recordings = []
    valid_supervisions = []
    for valid_path in valid_paths:
        audio_path = part_path / valid_path
        audio_path = audio_path.resolve()
        valid_path_splits = valid_path.split("/")
        audio_file_name = valid_path.replace("/", "_").replace(".wav", "")
        audio_file_name_splits = valid_path_splits[1].split("_")

        if valid_path_splits[0] == BACKGROUND_NOISE:
            speaker = None
            text = ""
        else:
            speaker = audio_file_name_splits[0]
            text = valid_path_splits[0].strip()

        if not audio_path.is_file():
            logging.warning(f"No such file: {audio_path}")
            continue
        valid_recording = Recording.from_file(
            path=audio_path,
            recording_id=audio_file_name,
        )
        valid_recordings.append(valid_recording)
        valid_segment = SupervisionSegment(
            id=audio_file_name,
            recording_id=audio_file_name,
            start=0.0,
            duration=valid_recording.duration,
            channel=0,
            language="English",
            speaker=speaker,
            text=text,
        )
        valid_supervisions.append(valid_segment)
    valid_recording_set = RecordingSet.from_recordings(valid_recordings)
    valid_supervision_set = SupervisionSet.from_segments(valid_supervisions)

    # Fix manifests
    valid_recording_set, valid_supervision_set = fix_manifests(
        valid_recording_set, valid_supervision_set
    )
    validate_recordings_and_supervisions(valid_recording_set, valid_supervision_set)

    return valid_recording_set, valid_supervision_set


def _prepare_test(
    speechcommands_version: str,
    corpus_dir: Pathlike,
) -> Tuple[RecordingSet, SupervisionSet]:
    """
    Returns the RecodingSet and SupervisionSet given a dataset part.
    :param speechcommands_version: str, dataset version.
    :param corpus_dir: Pathlike, the path of the data dir.
    :return: the RecodingSet and SupervisionSet for train and valid.
    """
    corpus_dir = Path(corpus_dir)
    part_path = corpus_dir / f"speech_commands_test_set_v0.0{speechcommands_version}"
    test_paths = []
    for root in os.listdir(part_path):
        if os.path.isdir(part_path / root):
            file_paths = os.listdir(part_path / root)
            test_paths += [
                os.path.join(root, file_path)
                for file_path in file_paths
                if file_path.endswith(".wav")
            ]

    test_paths = set(test_paths)

    test_recordings = []
    test_supervisions = []
    for test_path in test_paths:
        audio_path = part_path / test_path
        audio_path = audio_path.resolve()
        test_path_splits = test_path.split("/")
        audio_file_name = test_path.replace("/", "_").replace(".wav", "")
        audio_file_name_splits = test_path_splits[1].split("_")

        if test_path_splits[0] in WORDS:
            speaker = audio_file_name_splits[0]
            text = test_path_splits[0].strip()
        elif test_path_splits[0] == SILENCE:
            speaker = None
            text = ""
        elif test_path_splits[0] == UNKNOWN:
            speaker = audio_file_name_splits[1]
            text = audio_file_name_splits[0].strip()

        if not audio_path.is_file():
            logging.warning(f"No such file: {audio_path}")
            continue

        test_recording = Recording.from_file(
            path=audio_path,
            recording_id=audio_file_name,
        )
        test_recordings.append(test_recording)
        test_segment = SupervisionSegment(
            id=audio_file_name,
            recording_id=audio_file_name,
            start=0.0,
            duration=test_recording.duration,
            channel=0,
            language="English",
            speaker=speaker,
            text=text,
        )
        test_supervisions.append(test_segment)
    test_recording_set = RecordingSet.from_recordings(test_recordings)
    test_supervision_set = SupervisionSet.from_segments(test_supervisions)

    # Fix manifests
    test_recording_set, test_supervision_set = fix_manifests(
        test_recording_set, test_supervision_set
    )
    validate_recordings_and_supervisions(test_recording_set, test_supervision_set)

    return test_recording_set, test_supervision_set


def prepare_speechcommands(
    speechcommands_version: str,
    corpus_dir: Pathlike,
    output_dir: Optional[Pathlike] = None,
) -> Dict[str, Dict[str, Union[RecordingSet, SupervisionSet]]]:
    """
    Returns the manifests which consist of the Recordings and Supervisions
    :param speechcommands_version: str, dataset version.
    :param corpus_dir: Path to the Speech Commands dataset.
    :param output_dir: Pathlike, the path where to write the manifests.
    :return: a Dict whose key is the dataset part, and the value is Dicts with the keys 'recordings' and 'supervisions'.
    """
    corpus_dir = Path(corpus_dir)

    assert corpus_dir.is_dir(), f"No such directory: {corpus_dir}"

    logging.info("Preparing Speech Commands v0.0{speechcommands_version}...")
    dataset_parts = (
        f"speech_commands_v0.0{speechcommands_version}",
        f"speech_commands_test_set_v0.0{speechcommands_version}",
    )
    subsets = _SPLITS

    if output_dir is not None:
        output_dir = Path(output_dir)
        output_dir.mkdir(parents=True, exist_ok=True)

    manifests = defaultdict(dict)

    for part in tqdm(subsets, desc="Dataset parts"):
        logging.info(
            f"Processing Speech Commands v0.0{speechcommands_version} subset: {part}"
        )
        if manifests_exist(
            part=part,
            output_dir=output_dir,
            prefix=f"speechcommands{speechcommands_version}",
            suffix="jsonl.gz",
        ):
            logging.info(
                f"Speech Commands v0.0{speechcommands_version} subset: {part} already prepared - skipping."
            )
            continue
        if part == "train":
            train_paths, valid_paths = _split_train_valid(
                speechcommands_version, corpus_dir
            )
            recording_set, supervision_set = _prepare_train(
                speechcommands_version, train_paths, corpus_dir
            )
        elif part == "valid":
            recording_set, supervision_set = _prepare_valid(
                speechcommands_version, valid_paths, corpus_dir
            )
        elif part == "test":
            recording_set, supervision_set = _prepare_test(
                speechcommands_version, corpus_dir
            )

        if output_dir is not None:
            supervision_set.to_file(
                output_dir
                / f"speechcommands{speechcommands_version}_supervisions_{part}.jsonl.gz"
            )
            recording_set.to_file(
                output_dir
                / f"speechcommands{speechcommands_version}_recordings_{part}.jsonl.gz"
            )

        manifests[part] = {"recordings": recording_set, "supervisions": supervision_set}

    return manifests



"""
Official description from the "about" page of the Mozilla CommonVoice project
(source link: https://commonvoice.mozilla.org/en/about)

Why Common Voice?
Mozilla Common Voice is an initiative to help teach machines how real people speak.
This project is an effort to bridge the digital speech divide. Voice recognition technologies bring a human dimension to our devices, but developers need an enormous amount of voice data to build them. Currently, most of that data is expensive and proprietary. We want to make voice data freely and publicly available, and make sure the data represents the diversity of real people. Together we can make voice recognition better for everyone.

How does it work?
We are crowdsourcing an open-source dataset of voices. Donate your voice, validate the accuracy of other people's clips, make the dataset better for everyone.
"""
import csv
import logging
import math
import numbers
import shutil
import tarfile
import warnings
from collections import defaultdict
from concurrent.futures.process import ProcessPoolExecutor
from contextlib import contextmanager
from multiprocessing import get_context as mp_get_context
from pathlib import Path
from typing import Dict, Iterable, List, Optional, Sequence, Tuple, Union

from tqdm.auto import tqdm

from lhotse import (
    get_ffmpeg_torchaudio_info_enabled,
    load_manifest,
    set_ffmpeg_torchaudio_info_enabled,
    validate_recordings_and_supervisions,
)
from lhotse.audio import Recording, RecordingSet
from lhotse.qa import fix_manifests
from lhotse.supervision import SupervisionSegment, SupervisionSet
from lhotse.utils import Pathlike, is_module_available, resumable_download, safe_extract

DEFAULT_COMMONVOICE_URL = (
    "https://mozilla-common-voice-datasets.s3.dualstack.us-west-2.amazonaws.com"
)


COMMONVOICE_LANGS = "en de fr cy tt kab ca zh-TW it fa eu es ru tr nl eo zh-CN rw pt zh-HK cs pl uk".split()
COMMONVOICE_SPLITS = ("train", "dev", "test", "validated", "invalidated", "other")
COMMONVOICE_DEFAULT_SPLITS = ("test", "dev", "train")


def download_commonvoice(
    target_dir: Pathlike = ".",
    languages: Union[str, Iterable[str]] = "all",
    force_download: bool = False,
    base_url: str = DEFAULT_COMMONVOICE_URL,
    release: Optional[str] = "cv-corpus-13.0-2023-03-09",
) -> None:
    """
    Download and untar the CommonVoice dataset.

    :param target_dir: Pathlike, the path of the dir to storage the dataset.
    :param languages: one of: 'all' (downloads all known languages); a single language code (e.g., 'en'),
        or a list of language codes.
    :param force_download: Bool, if True, download the tars no matter if the tars exist.
    :param base_url: str, the base URL for CommonVoice.
    :param release: str, the name of the CommonVoice release (e.g., "cv-corpus-13.0-2023-03-09").
        It is used as part of the download URL.
    """
    # note(pzelasko): This code should work in general if we supply the right URL,
    # but the URL stopped working during the development of this script --
    # I'm not going to fight this, maybe somebody else would be interested to pick it up.
    target_dir = Path(target_dir)
    target_dir.mkdir(parents=True, exist_ok=True)
    url = f"{base_url}/{release}"

    if languages == "all":
        languages = COMMONVOICE_LANGS
    elif isinstance(languages, str):
        languages = [languages]
    else:
        languages = list(languages)

    logging.info(
        f"About to download {len(languages)} CommonVoice languages: {languages}"
    )
    for lang in tqdm(languages, desc="Downloading CommonVoice languages"):
        logging.info(f"Language: {lang}")
        # Split directory exists and seem valid? Skip this split.
        part_dir = target_dir / release / lang
        completed_detector = part_dir / ".completed"
        if completed_detector.is_file():
            logging.info(f"Skipping {lang} because {completed_detector} exists.")
            continue
        # Maybe-download the archive.
        tar_name = f"{lang}.tar.gz"
        tar_path = target_dir / tar_name
        if force_download or not tar_path.is_file():
            # After version 7.0, the commonvoice download address has changed
            if float(release.split("-")[2]) < 8.0:
                raise NotImplementedError(
                    "When the version is less than 8.0, CommonVoice requires you to enter e-mail to download the data.\n"
                    "Please download it manually for now.\n"
                    "Or you can choose a version greater than 8.0.\n"
                )
            else:
                # https://mozilla-common-voice-datasets.s3.dualstack.us-west-2.amazonaws.com/cv-corpus-13.0-2023-03-09/cv-corpus-13.0-2023-03-09-zh-CN.tar.gz
                single_url = url + f"/{release}-{lang}.tar.gz"
            resumable_download(
                single_url, filename=tar_path, force_download=force_download
            )
            logging.info(f"Downloading finished: {lang}")
        # Remove partial unpacked files, if any, and unpack everything.
        logging.info(f"Unpacking archive: {lang}")
        shutil.rmtree(part_dir, ignore_errors=True)
        with tarfile.open(tar_path) as tar:
            safe_extract(tar, path=target_dir)
        completed_detector.touch()


@contextmanager
def disable_ffmpeg_torchaudio_info() -> None:
    enabled = get_ffmpeg_torchaudio_info_enabled()
    set_ffmpeg_torchaudio_info_enabled(False)
    try:
        yield
    finally:
        set_ffmpeg_torchaudio_info_enabled(enabled)


def _read_cv_manifests_if_cached(
    output_dir: Optional[Pathlike],
    language: str,
) -> Dict[str, Dict[str, Union[RecordingSet, SupervisionSet]]]:
    """
    Returns:
        {'train': {'recordings': ..., 'supervisions': ...}, 'dev': ..., 'test': ...}
    """
    if output_dir is None:
        return {}
    manifests = defaultdict(dict)
    for part in ["train", "dev", "test"]:
        for manifest in ["recordings", "supervisions"]:
            path = output_dir / f"cv_{manifest}_{language}_{part}.jsonl.gz"
            if not path.is_file():
                continue
            manifests[part][manifest] = load_manifest(path)
    return manifests


def _parse_utterance(
    lang_path: Path,
    language: str,
    audio_info: str,
) -> Optional[Tuple[Recording, SupervisionSegment]]:
    audio_path = lang_path / "clips" / audio_info["path"]

    if not audio_path.is_file():
        logging.info(f"No such file: {audio_path}")
        return None

    recording_id = Path(audio_info["path"]).stem
    recording = Recording.from_file(path=audio_path, recording_id=recording_id)

    segment = SupervisionSegment(
        id=recording_id,
        recording_id=recording_id,
        start=0.0,
        duration=recording.duration,
        channel=0,
        language=language,
        speaker=audio_info["client_id"],
        text=audio_info["sentence"].strip(),
        gender=audio_info["gender"],
        custom={
            "age": audio_info["age"],
            "accents": audio_info["accents"],
            "variant": audio_info["variant"],
        },
    )
    return recording, segment


def _prepare_part(
    lang: str,
    part: str,
    lang_path: Pathlike,
    num_jobs: int = 1,
) -> Tuple[RecordingSet, SupervisionSet]:
    """
    Prepares part of CommonVoice data.

    :param lang: string language code (e.g., "en").
    :param part: which split to prepare (e.g., "train", "validated", etc.).
    :param lang_path: path to a CommonVoice directory for a specific language
        (e.g., "/path/to/cv-corpus-13.0-2023-03-09/pl").
    :param num_jobs: How many concurrent workers to use for scanning of the audio files.
    :return: a tuple of (RecordingSet, SupervisionSet) objects,
        note that CommonVoice manifests may be fairly large in memory.
    """

    lang_path = Path(lang_path)
    tsv_path = lang_path / f"{part}.tsv"

    with disable_ffmpeg_torchaudio_info():
        with ProcessPoolExecutor(
            max_workers=num_jobs,
            mp_context=mp_get_context("spawn"),
        ) as ex:

            futures = []
            recordings = []
            supervisions = []
            audio_infos = []

            with open(tsv_path, "r") as f:

                # Note: using QUOTE_NONE as CV dataset contains unbalanced quotes, cleanup needed later
                audio_infos = csv.DictReader(f, delimiter="\t", quoting=csv.QUOTE_NONE)

                for audio_info in tqdm(audio_infos, desc="Distributing tasks"):
                    futures.append(
                        ex.submit(
                            _parse_utterance,
                            lang_path,
                            lang,
                            audio_info,
                        )
                    )

            for future in tqdm(futures, desc="Processing"):
                result = future.result()
                if result is None:
                    continue
                recording, segment = result
                recordings.append(recording)
                supervisions.append(segment)

    recording_set = RecordingSet.from_recordings(recordings)
    supervision_set = SupervisionSet.from_segments(supervisions)

    return recording_set, supervision_set


def prepare_commonvoice(
    corpus_dir: Pathlike,
    output_dir: Pathlike,
    languages: Union[str, Sequence[str]] = "auto",
    splits: Union[str, Sequence[str]] = COMMONVOICE_DEFAULT_SPLITS,
    num_jobs: int = 1,
) -> Dict[str, Dict[str, Dict[str, Union[RecordingSet, SupervisionSet]]]]:
    """
    Returns the manifests which consist of the Recordings and Supervisions.
    When all the manifests are available in the ``output_dir``, it will simply read and return them.

    This function expects the input directory structure of::

        >>> metadata_path = corpus_dir / language_code / "{train,dev,test}.tsv"
        >>> # e.g. pl_train_metadata_path = "/path/to/cv-corpus-13.0-2023-03-09/pl/train.tsv"
        >>> audio_path = corpus_dir / language_code / "clips"
        >>> # e.g. pl_audio_path = "/path/to/cv-corpus-13.0-2023-03-09/pl/clips"

    Returns a dict with 3-level structure (lang -> split -> manifest-type)::

        >>> {'en/fr/pl/...': {'train/dev/test': {'recordings/supervisions': manifest}}}

    :param corpus_dir: Pathlike, the path to the downloaded corpus.
    :param output_dir: Pathlike, the path where to write the manifests.
    :param languages: 'auto' (prepare all discovered data) or a list of language codes.
    :param splits: by default ``['train', 'dev', 'test']``, can also include
        ``'validated'``, ``'invalidated'``, and ``'other'``.
    :param num_jobs: How many concurrent workers to use for scanning of the audio files.
    :return: a dict with manifests for all specified languagues and their train/dev/test splits.
    """
    corpus_dir = Path(corpus_dir)
    assert corpus_dir.is_dir(), f"No such directory: {corpus_dir}"
    assert output_dir is not None, (
        "CommonVoice recipe requires to specify the output "
        "manifest directory (output_dir cannot be None)."
    )
    output_dir = Path(output_dir)
    output_dir.mkdir(parents=True, exist_ok=True)

    if languages == "auto":
        languages = set(COMMONVOICE_LANGS).intersection(
            path.name for path in corpus_dir.glob("*")
        )
        if not languages:
            raise ValueError(
                f"Could not find any of CommonVoice languages in: {corpus_dir}"
            )
    elif isinstance(languages, str):
        languages = [languages]

    manifests = {}

    for lang in tqdm(languages, desc="Processing CommonVoice languages"):
        logging.info(f"Language: {lang}")
        lang_path = corpus_dir / lang

        # Maybe the manifests already exist: we can read them and save a bit of preparation time.
        # Pattern: "cv_recordings_en_train.jsonl.gz" / "cv_supervisions_en_train.jsonl.gz"
        lang_manifests = _read_cv_manifests_if_cached(
            output_dir=output_dir, language=lang
        )

        for part in tqdm(splits, desc=f"Spliting"):
            logging.info(f"Spliting {part}")
            if part in lang_manifests:
                logging.info(
                    f"{part} split of CommonVoice-{lang} already prepared - skipping."
                )
                continue
            recording_set, supervision_set = _prepare_part(
                lang=lang,
                part=part,
                lang_path=lang_path,
                num_jobs=num_jobs,
            )

            # Fix manifests
            recording_set, supervision_set = fix_manifests(
                recording_set, supervision_set
            )
            validate_recordings_and_supervisions(recording_set, supervision_set)

            supervision_set.to_file(
                output_dir / f"cv-{lang}_supervisions_{part}.jsonl.gz"
            )
            recording_set.to_file(output_dir / f"cv-{lang}_recordings_{part}.jsonl.gz")

            lang_manifests[part] = {
                "supervisions": supervision_set,
                "recordings": recording_set,
            }

        manifests[lang] = lang_manifests

    return manifests



from functools import lru_cache, wraps
from threading import Lock
from typing import Any, Callable, Dict, Optional

LHOTSE_CACHING_ENABLED = False

# This dict is holding two variants for each registered method:
# the cached variant, that uses its own LRU cache,
# and a noncached variant, that will always be recomputed.
# This allows the users to enable or disable caching and
# always get the expected behavior.
LHOTSE_CACHED_METHOD_REGISTRY = {"cached": {}, "noncached": {}}


def set_caching_enabled(enabled: bool) -> None:
    global LHOTSE_CACHING_ENABLED
    global LHOTSE_CACHED_METHOD_REGISTRY
    assert isinstance(enabled, bool)
    LHOTSE_CACHING_ENABLED = enabled

    # enable cache for audio files of "command" type
    AudioCache.enable(enabled)

    if not enabled:
        # Caching disabled: purge all caches.
        for method in LHOTSE_CACHED_METHOD_REGISTRY["cached"].values():
            method.cache_clear()


def is_caching_enabled() -> bool:
    return LHOTSE_CACHING_ENABLED


def dynamic_lru_cache(method: Callable) -> Callable:
    """
    Least-recently-used cache decorator.

    It enhances Python's built-in ``lru_cache`` with a dynamic
    lookup of whether to apply the cached, or noncached variant
    of the decorated function.

    To disable/enable caching globally in Lhotse, call::

        >>> from lhotse import set_caching_enabled
        >>> set_caching_enabled(True)   # enable
        >>> set_caching_enabled(False)  # disable

    Currently it hard-codes the cache size at 512 items
    (regardless of the array size).

    Check :meth:`functools.lru_cache` for more details.
    """
    # Create an LRU cached variant of the method and register it
    # together with the original, noncached method.
    global LHOTSE_CACHED_METHOD_REGISTRY
    name = method.__qualname__  # example: "Recording.load_audio()"
    if name in LHOTSE_CACHED_METHOD_REGISTRY["cached"]:
        raise ValueError(
            f"Method '{name}' is already cached. "
            f"We don't support caching different methods which have "
            f"the same __qualname__ attribute (i.e., class name + method name)."
        )
    LHOTSE_CACHED_METHOD_REGISTRY["noncached"][name] = method
    LHOTSE_CACHED_METHOD_REGISTRY["cached"][name] = lru_cache(maxsize=512)(method)

    @wraps(method)
    def wrapper(*args, **kwargs) -> Any:
        # Each time the user calls this function, we will
        # dynamically dispatch the cached or noncached variant
        # of the wrapped method, depending on the global settings.
        if is_caching_enabled():
            m = LHOTSE_CACHED_METHOD_REGISTRY["cached"][name]
        else:
            m = LHOTSE_CACHED_METHOD_REGISTRY["noncached"][name]
        return m(*args, **kwargs)

    return wrapper


class AudioCache:
    """
    Cache of 'bytes' objects with audio data.
    It is used to cache the "command" type audio inputs.

    By default it is disabled, to enable call `set_caching_enabled(True)`
    or `AudioCache.enable()`.

    The cache size is limited to max 100 elements and 500MB of audio.

    A global dict `__cache_dict` (static member variable of class AudioCache)
    is holding the wavs as 'bytes' arrays.
    The key is the 'source' identifier (i.e. the command for loading the data).

    Thread-safety is ensured by a threading.Lock guard.
    """

    __enabled: bool = False

    max_cache_memory: int = 500 * 1e6  # 500 MB
    max_cache_elements: int = 100  # 100 audio files

    __cache_dict: Dict[str, bytes] = {}
    __lock: Lock = Lock()

    @classmethod
    def enable(cls, enabled=True):
        cls.__enabled = enabled
        if not enabled:
            cls.__clear_cache()

    @classmethod
    def enabled(cls) -> bool:
        return cls.__enabled

    @classmethod
    def try_cache(cls, key: str) -> Optional[bytes]:
        """
        Test if 'key' is in the chache. If yes return the bytes array,
        otherwise return None.
        """

        if not cls.__enabled:
            return None

        with cls.__lock:
            if key in cls.__cache_dict:
                return cls.__cache_dict[key]
            else:
                return None

    @classmethod
    def add_to_cache(cls, key: str, value: bytes):
        """
        Add the new (key,value) pair to cache.
        Possibly free some elements before adding the new pair.
        The oldest elements are removed first.
        """

        if not cls.__enabled:
            return None

        if len(value) > cls.max_cache_memory:
            return

        with cls.__lock:
            # limit cache elements
            while len(cls.__cache_dict) > cls.max_cache_elements:
                # remove oldest elements from cache
                # (dict pairs are sorted according to insertion order)
                cls.__cache_dict.pop(next(iter(cls.__cache_dict)))

            # limit cache memory
            while len(value) + AudioCache.__cache_memory() > cls.max_cache_memory:
                # remove oldest elements from cache
                # (dict pairs are sorted according to insertion order)
                cls.__cache_dict.pop(next(iter(cls.__cache_dict)))

            # store the new (key,value) pair
            cls.__cache_dict[key] = value

    @classmethod
    def __cache_memory(cls) -> int:
        """
        Return size of AudioCache values in bytes.
        (internal, not to be called from outside)
        """
        ans = 0
        for key, value in cls.__cache_dict.items():
            ans += len(value)
        return ans

    @classmethod
    def __clear_cache(cls) -> None:
        """
        Clear the cache, remove the data.
        """
        with cls.__lock:
            cls.__cache_dict.clear()



from dataclasses import asdict, dataclass
from typing import Any, Dict

import numpy as np

from lhotse.features.base import TorchaudioFeatureExtractor, register_extractor
from lhotse.utils import EPSILON, Seconds


@dataclass
class TorchaudioSpectrogramConfig:
    # Note that `snip_edges` parameter is missing from config: in order to simplify the relationship between
    #  the duration and the number of frames, we are always setting `snip_edges` to False.
    dither: float = 0.0
    window_type: str = "povey"
    # Note that frame_length and frame_shift will be converted to milliseconds before torchaudio/Kaldi sees them
    frame_length: Seconds = 0.025
    frame_shift: Seconds = 0.01
    remove_dc_offset: bool = True
    round_to_power_of_two: bool = True
    energy_floor: float = EPSILON
    min_duration: float = 0.0
    preemphasis_coefficient: float = 0.97
    raw_energy: bool = True

    def to_dict(self) -> Dict[str, Any]:
        return asdict(self)

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> "TorchaudioSpectrogramConfig":
        return TorchaudioSpectrogramConfig(**data)


@register_extractor
class TorchaudioSpectrogram(TorchaudioFeatureExtractor):
    """Log spectrogram feature extractor based on ``torchaudio.compliance.kaldi.spectrogram`` function."""

    name = "spectrogram"
    config_type = TorchaudioSpectrogramConfig

    def _feature_fn(self, *args, **kwargs):
        from torchaudio.compliance.kaldi import spectrogram

        return spectrogram(*args, **kwargs)

    def feature_dim(self, sampling_rate: int) -> int:
        from torchaudio.compliance.kaldi import _next_power_of_2

        window_size = int(self.config.frame_length * sampling_rate)
        return (
            _next_power_of_2(window_size) // 2 + 1
            if self.config.round_to_power_of_two
            else window_size
        )

    @staticmethod
    def mix(
        features_a: np.ndarray, features_b: np.ndarray, energy_scaling_factor_b: float
    ) -> np.ndarray:
        # Torchaudio returns log-power spectrum, hence the need for logsumexp
        return np.log(
            np.maximum(
                # protection against log(0); max with EPSILON is adequate since these are energies (always >= 0)
                EPSILON,
                np.exp(features_a) + energy_scaling_factor_b * np.exp(features_b),
            )
        )

    @staticmethod
    def compute_energy(features: np.ndarray) -> float:
        # Torchaudio returns log-power spectrum, hence the need for exp before the sum
        return float(np.sum(np.exp(features)))



"""
About the MuST-C corpus

MuST-C is a multilingual speech translation corpus whose size and quality will
facilitate the training of end-to-end systems for SLT from English into a set
of languages.

For each target language, MuST-C comprises several hundred hours of audio
recordings from English TED Talks, which are automatically aligned at the
sentence level with their manual transcriptions and translations.

We don't provide download_mustc().

Please refer to
https://ict.fbk.eu/must-c-releases/
for downloading.

If you have downloaded and extracted the dataset to the directory

/ceph-data3/fangjun/data/must-c/v2.0/en-de
/ceph-data3/fangjun/data/must-c/v2.0/en-zh

You can call lhotse with the following commands

(1) When the target language is German:

    lhotse prepare must-c \
      --tgt-lang de \
      -j 10 \
      /ceph-data3/fangjun/data/must-c/v2.0/ \
      ./data/manifests/v2.0

(2) When the target language is Chinese:

    lhotse prepare must-c \
      --tgt-lang zh \
      -j 10 \
      /ceph-data3/fangjun/data/must-c/v2.0/ \
      ./data/manifests/v2.0
"""

import logging
from itertools import groupby, repeat
from pathlib import Path
from typing import Dict, List, Tuple, Union

from tqdm.auto import tqdm

from lhotse import fix_manifests, validate_recordings_and_supervisions
from lhotse.audio import Recording, RecordingSet
from lhotse.parallel import parallel_map
from lhotse.serialization import load_yaml
from lhotse.supervision import SupervisionSegment, SupervisionSet
from lhotse.utils import Pathlike, Seconds


def prepare_must_c(
    corpus_dir: Pathlike,
    output_dir: Pathlike,
    tgt_lang: str,
    num_jobs: int = 1,
) -> Dict[str, Dict[str, Union[RecordingSet, SupervisionSet]]]:
    """Prepare manifests for the MUST-C corpus.

    :param: corpus_dir: We assume there is a folder {src_lang}-{tgt_lang} inside
        this directory.
    :param: output_dir: Directory where the manifests should be written.
    :param: tgt_lang: Target language, e.g., zh, de, etc.
    :param: src_lang: Source language, e.g., en.
    :param: num_jobs: Number of processes to use for parsing the data
    """
    src_lang = "en"

    in_data_dir = Path(corpus_dir) / f"{src_lang}-{tgt_lang}/data"
    assert in_data_dir.is_dir(), in_data_dir

    datasets = ["dev", "tst-COMMON", "tst-HE", "train"]
    output_dir = Path(output_dir)
    output_dir.mkdir(parents=True, exist_ok=True)

    for d in datasets:
        logging.info(f"Processing {d}")
        dataset_dir = in_data_dir / d
        # In the dataset_dir, we can find two directories: txt and wav
        # In the txt directory, we can find the following files:
        #   {d}.{src_lang}, {d}.{tgt_lang}, {d}.yaml
        assert dataset_dir.is_dir(), dataset_dir / d

        with open(dataset_dir / "txt" / f"{d}.{tgt_lang}") as f:
            transcripts = [line.strip() for line in f]

        segments = load_yaml(dataset_dir / "txt" / f"{d}.yaml")
        assert len(transcripts) == len(segments), (len(transcripts), len(segments))

        # segments[0] contains something like below:
        #  {'duration': 3.5, 'offset': 16.08, 'rW': 9, 'uW': 0,
        #  'speaker_id': 'spk.767', 'wav': 'ted_767.wav'}
        groups = []
        start = 0
        for _, group in groupby(segments, lambda x: x["wav"]):
            this_wave = [list(group)]
            num_sentences = len(this_wave[0])
            end = start + num_sentences

            this_wave.append(transcripts[start:end])
            start = end

            groups.append(this_wave)

        assert start == len(transcripts), (start, len(transcripts))

        recording_list = []
        supervision_list = []
        for recording, sup_segments in tqdm(
            parallel_map(
                parse_utterance,
                repeat(dataset_dir / "wav"),
                groups,
                repeat(tgt_lang),
                num_jobs=num_jobs,
            ),
            desc=f"Processing must-c {d}",
        ):
            recording_list.append(recording)
            supervision_list.extend(sup_segments)

        recordings, supervisions = fix_manifests(
            recordings=RecordingSet.from_recordings(recording_list),
            supervisions=SupervisionSet.from_segments(supervision_list),
        )

        validate_recordings_and_supervisions(
            recordings=recordings, supervisions=supervisions
        )

        with RecordingSet.open_writer(
            output_dir / f"must_c_recordings_{src_lang}-{tgt_lang}_{d}.jsonl.gz"
        ) as rec_writer:
            for r in recordings:
                rec_writer.write(r)

        with SupervisionSet.open_writer(
            output_dir / f"must_c_supervisions_{src_lang}-{tgt_lang}_{d}.jsonl.gz"
        ) as sup_writer:
            for sup in supervisions:
                sup_writer.write(sup)


def parse_utterance(
    wave_dir: Path,
    groups: Tuple[List[dict], List[str]],
    tgt_lang: str,
) -> Tuple[Recording, List[SupervisionSegment]]:
    """
    :param: wave_dir: The wave directory. It contains *.wav files.
    :param: groups: A tuple containing two lists. The first one is a list
        of dict, where each dict contains something like below:

          {duration: 3.500000, offset: 16.080000, rW: 9, uW: 0,
           speaker_id: spk.767, wav: ted_767.wav}

        The second one is a list of transcripts.
    :param: tgt_lang: The language of the transcript, e.g., zh, en, de, etc.
    """
    wave_segments = groups[0]
    transcripts = groups[1]

    assert len(wave_segments) == len(transcripts), (
        len(wave_segments),
        len(transcripts),
    )

    wave_file = wave_dir / wave_segments[0]["wav"]
    recording = Recording.from_file(wave_file)

    segments = []
    for i, (wave_segment, transcript) in enumerate(zip(wave_segments, transcripts)):
        segments.append(
            SupervisionSegment(
                id=f"{recording.id}-seg-{i}",
                recording_id=recording.id,
                start=Seconds(wave_segment["offset"]),
                duration=round(Seconds(wave_segment["duration"]), ndigits=8),
                channel=0,
                language=tgt_lang,
                speaker=wave_segment["speaker_id"],
                text=transcript,
            )
        )
    return recording, segments



from typing import List

import click

from lhotse.bin.modes import download, prepare
from lhotse.recipes.tedlium import TEDLIUM_PARTS, download_tedlium, prepare_tedlium
from lhotse.utils import Pathlike


@prepare.command()
@click.argument(
    "tedlium_dir", type=click.Path(exists=True, dir_okay=True, file_okay=False)
)
@click.argument("output_dir", type=click.Path())
@click.option(
    "--parts",
    "-p",
    type=click.Choice(TEDLIUM_PARTS),
    multiple=True,
    default=list(TEDLIUM_PARTS),
    help="Which parts of TED-LIUM v3 to prepare (by default all).",
)
@click.option(
    "-j",
    "--num-jobs",
    type=int,
    default=1,
    help="How many threads to use (can give good speed-ups with slow disks).",
)
@click.option(
    "--normalize-text",
    type=click.Choice(["none", "upper", "kaldi"], case_sensitive=False),
    default="none",
    help="Type of text normalization to apply (no normalization, by default). "
    "Selecting `kaldi` will remove <unk> tokens and join suffixes.",
)
def tedlium(
    tedlium_dir: Pathlike,
    output_dir: Pathlike,
    parts: List[str],
    num_jobs: int,
    normalize_text: str,
):
    """
    TED-LIUM v3 recording and supervision manifest preparation.
    """
    prepare_tedlium(
        tedlium_root=tedlium_dir,
        output_dir=output_dir,
        dataset_parts=parts,
        num_jobs=num_jobs,
        normalize_text=normalize_text,
    )


@download.command()
@click.argument("target_dir", type=click.Path())
def tedlium(target_dir: Pathlike):
    """TED-LIUM v3 download (approx. 11GB)."""
    download_tedlium(target_dir)
